"""
The core game package
"""

from texasholdem.game.action_type import ActionType
from texasholdem.game.game import TexasHoldEm, GameState
from texasholdem.game.hand_phase import HandPhase
from texasholdem.game.history import (History,
                                      FILE_EXTENSION,
                                      HistoryImportError,
                                      SettleHistory,
                                      BettingRoundHistory,
                                      PrehandHistory,
                                      PlayerAction)
from texasholdem.game.player_state import PlayerState



import random
from typing import List

from texasholdem.card import card
from texasholdem.card.card import Card


class Deck:
    """
    Class representing a deck. The first time we create, we seed the static
    deck with the list of unique card integers. Each object instantiated simply
    makes a copy of this object and shuffles it.

    """
    _FULL_DECK: List[Card] = []

    def __init__(self):
        self.cards = Deck._get_full_deck()
        self.shuffle()

    def shuffle(self) -> None:
        """
        Shuffles the remaining cards in the deck.

        """
        random.shuffle(self.cards)

    def draw(self, num=1) -> List[Card]:
        """
        Draw card(s) from the deck. These cards leave the deck and are not saved.

        Args:
            num (int): How many cards to draw. Defaults to 1.
        Returns:
           List[Card]: A list of length n of cards (See :class:`~texasholdem.card.card.Card`).
        Raises:
            ValueError: If the deck size is less than the given n.

        """
        if len(self.cards) < num:
            raise ValueError(f'Cannot draw {num} cards from deck of size {len(self.cards)}')

        cards = self.cards[:num]
        self.cards = self.cards[num:]
        return cards

    def __str__(self) -> str:
        return card.card_list_to_pretty_str(self.cards)

    @staticmethod
    def _get_full_deck() -> List[Card]:
        if Deck._FULL_DECK:
            return list(Deck._FULL_DECK)

        # create the standard 52 card deck
        for rank in Card.STR_RANKS:
            for suit in Card.CHAR_SUIT_TO_INT_SUIT:
                Deck._FULL_DECK.append(Card(rank + suit))

        return list(Deck._FULL_DECK)



from enum import Enum, auto


class ActionType(Enum):
    """An enum representing the types of actions a player can take."""

    RAISE = auto()
    """Raises the latest pot."""

    ALL_IN = auto()
    """Posts all available chips to the current pot."""

    CALL = auto()
    """Posts the minimum chips to remain in the pot."""

    CHECK = auto()
    """Passes the action."""

    FOLD = auto()
    """Folds the hand and exits the pot."""



from __future__ import annotations

import math
from typing import Union, Iterable, List


class Card(int):
    """
    Usage :code:`Card("Kd")`

    We represent :class:`~texasholdem.card.card.Card` objects as native Python 32-bit
    integers. Most of the bits are used, and have a specific meaning. See below:

    .. table:: Card
        :align: center
        :widths: auto

        ========  ========  ========  ========
        xxxbbbbb  bbbbbbbb  cdhsrrrr  xxpppppp
        ========  ========  ========  ========


    - p = prime number of rank (in binary) (deuce=2, trey=3, four=5, ..., ace=41)
    - r = rank of card (in binary) (deuce=0, trey=1, four=2, five=3, ..., ace=12)
    - cdhs = suit of card (bit turned on based on suit of card)
    - b = bit turned on depending on rank of card (deuce=1st bit, trey=2nd bit, ...)
    - x = unused

    **Example**
        .. table::
            :align: center
            :widths: auto

            ================ ========  ========  ========  ========
            Card             xxxAKQJT  98765432  CDHSrrrr  xxPPPPPP
            ================ ========  ========  ========  ========
            King of Diamonds 00001000  00000000  01001011  00100101
            Five of Spades   00000000  00001000  00010011  00000111
            Jack of Clubs    00000010  00000000  10001001  00011101
            ================ ========  ========  ========  ========


    This representation allows for minimal memory overhead along with fast applications
    necessary for poker:

        - Make a unique prime product for each hand (by multiplying the prime bits)
        - Detect flushes (bitwise && for the suits)
        - Detect straights (shift and bitwise &&)

    Args:
        arg (str | int): A string of the form "{rank}{suit}" e.g. "Kd" or "As" or a properly-formed
            Card-int as described above.

    """

    # the basics
    STR_RANKS = '23456789TJQKA'
    INT_RANKS = tuple(range(13))
    PRIMES = (2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41)

    # conversion from string => int
    CHAR_RANK_TO_INT_RANK = dict(zip(list(STR_RANKS), INT_RANKS))
    CHAR_SUIT_TO_INT_SUIT = {
        's': 1,  # spades
        'h': 2,  # hearts
        'd': 4,  # diamonds
        'c': 8,  # clubs
    }
    INT_SUIT_TO_CHAR_SUIT = 'xshxdxxxc'

    # for pretty printing
    PRETTY_SUITS = {
        1: "\u2660",  # spades
        2: "\u2665",  # hearts
        4: "\u2666",  # diamonds
        8: "\u2663"  # clubs
    }

    def __new__(cls, arg: Union[str, int]) -> Card:
        if isinstance(arg, str):
            return Card.from_string(arg)
        return Card.from_int(arg)

    @classmethod
    def from_string(cls, string: str) -> Card:
        """
        Converts Card string to binary integer representation of card, inspired by:

        http://www.suffecool.net/poker/evaluator.html

        Example:
            "Kd" --> int(0b 00001000  00000000  01001011  00100101) = 134236965

        Args:
            string (str): A string representing a card.
        Returns:
            Card: The 32-bit int representing the card as described above
        """

        rank_char = string[0]
        suit_char = string[1]
        rank_int = Card.CHAR_RANK_TO_INT_RANK[rank_char]
        suit_int = Card.CHAR_SUIT_TO_INT_SUIT[suit_char]
        rank_prime = Card.PRIMES[rank_int]

        bitrank = 1 << rank_int << 16
        suit = suit_int << 12
        rank = rank_int << 8

        card_int = bitrank | suit | rank | rank_prime
        return Card.from_int(card_int)

    @classmethod
    def from_int(cls, card_int: int) -> Card:
        """
        Converts an already well-formed card integer as described above

        Example:
            134236965 --> Card("Kd")

        Args:
            card_int (int): An int representing a card.
        Returns:
            Card: The 32-bit int representing the card as described above

        """
        return super(Card, cls).__new__(cls, card_int)

    def __str__(self) -> str:
        """
        Translates card into a readable string.

        Example:
            134236965 --> "Kd"

        Returns:
            str: The human-readable string representing this card.

        """
        return Card.STR_RANKS[self.rank] + Card.INT_SUIT_TO_CHAR_SUIT[self.suit]

    def __repr__(self) -> str:
        return f'Card("{str(self)}")'

    @property
    def rank(self) -> int:
        """
        The rank of the card as an int.

        Example:
            134236965 ("Kd") --> 11
            268440327 ("As") --> 12

        Returns:
            int: Number between 0-12, representing the rank of the card.

        """

        return (self >> 8) & 0xF

    @property
    def suit(self) -> int:
        """
        The suit int of the card using the following table:

        .. table::
            :align: center
            :widths: auto

            ========  ======
            Suit      Number
            ========  ======
            Spades      1
            Hearts      2
            Diamonds    4
            Clubs       8
            ========  ======

        Example:
            134236965 ("Kd") --> 2

        Returns:
            int: 1,2,4, or 8, representing the suit of the card from the above table.

        """
        return (self >> 12) & 0xF

    @property
    def bitrank(self) -> int:
        """
        The bitrank of the card. This returns 2^k where k is the
        rank of the card.

        Example:
            134236965 ("Kd") --> 2^11

        Returns:
            int: 2^k where k is the rank of the card.

        """
        return (self >> 16) & 0x1FFF

    @property
    def prime(self) -> int:
        """
        The prime associated with the card. This returns the kth prime
        starting at 2 where k is the rank of the card.

        Example:
            134236965 ("Kd") --> 37

        """
        return self & 0x3F

    @property
    def pretty_string(self) -> str:
        """
        A human-readable pretty string with ascii suits.

        """
        return f"[ {Card.STR_RANKS[self.rank]} {Card.PRETTY_SUITS[self.suit]} ]"

    @property
    def binary_string(self) -> str:
        """
        For debugging purposes. Displays the binary number as a
        human readable string in groups of four digits.

        """
        bstr = bin(self)[2:][::-1]  # chop off the 0b and THEN reverse string
        output = list("".join(["0000" + "\t"] * 7) + "0000")

        for i, b_char in enumerate(bstr, 0):
            output[i + int(i/4)] = b_char

        # output the string to console
        output.reverse()
        return "".join(output)


def card_strings_to_int(card_strs: Iterable[str]) -> List[Card]:
    """
    Args:
        card_strs (Iterable[str]): An iterable of card strings.
    Returns:
        List[Card]: The cards in the corresponding int format.

    """
    bhand = []
    for card_str in card_strs:
        bhand.append(Card(card_str))
    return bhand


def prime_product_from_hand(cards: Iterable[Card]) -> int:
    """
    Args:
        cards (Iterable[Card]): An Iterable of cards
    Returns:
        int: The product of all primes in the hand, corresponding to the rank of the
            card (See :meth:`Card.prime`)

    """
    return math.prod(card.prime for card in cards)


def prime_product_from_rankbits(rankbits: int) -> int:
    """
    Returns the prime product using the bitrank (b)
    bits of the hand. Each 1 in the sequence is converted
    to the correct prime and multiplied in.

    Primarily used for evaulating flushes and straights,
    two occasions where we know the ranks are *ALL* different.
    Assumes that the input is in form (set bits):

    .. table::
            :align: center
            :widths: auto

            ========  ========
            xxxbbbbb  bbbbbbbb
            ========  ========

    Args:
        rankbits (int): a single 32-bit (only 13-bits set) integer representing
            the ranks of 5 *different* ranked card (5 of 13 bits are set)
    Returns:
        int: The product of all primes in the hand, corresponding
            to the rank of the card.

    """
    product = 1
    for i in Card.INT_RANKS:
        # if the ith bit is set
        if rankbits & (1 << i):
            product *= Card.PRIMES[i]
    return product


def card_list_to_pretty_str(cards: List[Card]) -> str:
    """
    Prints the given card in a human-readable pretty string with
    ascii suit.

    Args:
        cards (List[Card]): A list of card ints in the proper form.
    Returns:
        string: A human-readable pretty string with ascii suits.

    """
    return " ".join(card.pretty_string for card in cards)



# pylint: disable=too-many-lines
"""
The game module includes lightweight data classes:

    - :class:`Player`
    - :class:`Pot`
    - :class:`GameState`

It also includes the main :class:`TexasHoldEm` class.

"""
from __future__ import annotations
import os
from typing import Iterator, Callable, Dict, Tuple, Optional, Union, List, Iterable
from collections import deque
from enum import Enum, auto
import random
import warnings

from deprecated.sphinx import versionadded, versionchanged

from texasholdem.card.card import Card
from texasholdem.card.deck import Deck
from texasholdem.game.history import (History, PrehandHistory,
                                      BettingRoundHistory, PlayerAction,
                                      HistoryImportError, SettleHistory)
from texasholdem.game.action_type import ActionType
from texasholdem.game.hand_phase import HandPhase
from texasholdem.game.player_state import PlayerState
from texasholdem.evaluator import evaluator
from texasholdem.util.functions import check_raise


class Player:
    # pylint: disable=too-few-public-methods
    """
    The :class:`TexasHoldEm` class uses this
    class as a bookkeeping mechanism.

    Attributes:
        player_id (int): The player id
        chips (int): The number of chips the player has behind them
        state (PlayerState): The player state
        last_pot (int): The pot id of the last pot the player is eligible for.

    """

    def __init__(self, player_id, chips):
        self.player_id = player_id
        self.chips = chips
        self.state = PlayerState.IN

        # invariant: last_pot is the newest pot that player is eligible for
        self.last_pot = 0


class Pot:
    """
    Class representing a betting pot, players will post to the pot.
    If players post more than the raised amount, we set the new raised amount.

    At the end of a betting round, the bets are consolidated and reset.

    Attributes:
        amount (int): The amount of chips in the pot NOT including the current betting round
        raised (int): The highest bet amount in the current betting round
        player_amounts (Dict[int, int]): A map from player id to # chips each player has posted
                                         this round

    """

    def __init__(self):
        self.amount = 0
        self.raised = 0
        self.player_amounts = {}

    def chips_to_call(self, player_id: int) -> int:
        """Returns the amount of chips to call for the given player.

        Args:
            player_id (int): The player_id of the player in the pot

        Returns:
            int: The amount the player needs to call to be in this pot.
              This is just :attr:`raised` if the player hasn't bet yet.

        """
        return self.raised - self.player_amounts.get(player_id, 0)

    def player_post(self, player_id: int, amount: int):
        """
        The given player posts amount into this pot. If the total amount for the player
        exceeds the current :attr:`raised` value, sets it anew.

        Arguments:
            player_id (int): The player_id of the player posting
            amount (int): The amount to post into this pot

        """
        self.player_amounts[player_id] = self.player_amounts.get(player_id, 0) + amount

        if self.player_amounts[player_id] > self.raised:
            self.raised = self.player_amounts[player_id]

    def get_player_amount(self, player_id: int) -> int:
        """
        Arguments:
            player_id (int): The player id
        Returns:
            int: the amount the player has bet currently for this pot.

        """
        return self.player_amounts.get(player_id, 0)

    def players_in_pot(self) -> Iterator[int]:
        """
        Returns:
            Iterator[int]: An iterator over the player player_id's that have a stake in this pot.

        """

        return iter(self.player_amounts.keys())

    def collect_bets(self):
        """
        Collects all the bets players made this round and adds them to
        the total amount. Resets player betting.

        """
        self.raised = 0
        for player_id, player_amount in self.player_amounts.items():
            self.amount += player_amount
            self.player_amounts[player_id] = 0

    def remove_player(self, player_id: int):
        """
        Removes the given player from the pot and adds their betting
        to the total amount.

        Arguments:
            player_id (int): The player id

        """
        if player_id not in self.player_amounts:
            return

        self.amount += self.player_amounts.pop(player_id)

    def get_amount(self) -> int:
        """
        Returns:
            int: How many chips this pot contains not including the current round of betting.

        """

        return self.amount

    def get_total_amount(self) -> int:
        """
        Returns:
            int: How many chips this pot contains including the current round of betting.

        """

        return sum(self.player_amounts.values()) + self.get_amount()

    def split_pot(self, raised_level: int) -> Optional[Pot]:
        """
        Returns a pot with players and the overflow over the raised_level

        Arguments:
            raised_level (int): The chip count to cut off at
        Returns:
            Pot, optional: The new Pot or None if self.raised <= raised_level

        """
        if self.raised <= raised_level:
            return None

        split_pot = Pot()
        self.raised = raised_level

        for player_id in self.players_in_pot():
            # player currently in last pot, post overflow to the split pot
            if self.get_player_amount(player_id) > self.raised:
                overflow = self.get_player_amount(player_id) - self.raised
                split_pot.player_post(player_id, overflow)
                self.player_amounts[player_id] -= overflow

        return split_pot


class GameState(Enum):
    """An enum representing the state of the game (not just one hand)"""

    RUNNING = auto()
    """The table is active and is able to play hands."""

    STOPPED = auto()
    """The table is inactive due to lack of players or lack
    of chips and unable to play hands."""


class TexasHoldEm:
    # pylint: disable=too-many-instance-attributes,stop-iteration-return

    """
    Represents a table of TexasHoldEm (tournament style).

    Instantiate this object with the buyin, big blind, small blind,
    and the number of players.

    To interact with this class, call :meth:`TexasHoldEm.start_hand` which will
    run the PREHAND phase (move/post blinds, reset pots, deal cards, etc.)

    To input an action at each stage, call :meth:`TexasHoldEm.take_action` which will
    execute the given action for the current player.

    Arguments:
        buyin (int): The buyin to register for this game.
        big_blind (int): Big blind
        small_blind (int): Small blind
        max_players (int): how many players can sit at the table, defaults to 9.
    Attributes:
        buyin (int): The buyin to register for this game.
        big_blind (int): Big blind
        small_blind (int): Small blind
        max_players (int): how many players can sit at the table, defaults to 9.
        players (List[Player]): A list of all players in the game.
        btn_loc (int): The id of the player who has the button.
        bb_loc (int): The id of the player who has the big blind.
        sb_loc (int): The id of the player who has the small blind.
        current_player (int): The id of the player who is to act.
        pots (List[Pot]): The active :class:`Pot` objects in the game.
        board (List[Card]): The communal cards that are out.
        hands (Dict[int, List[Card]]): Map from player id to their hand
        last_raise (int): The amount of the last raise (that was over the previous bet)
        raise_option (bool): If the current player has the option to raise (not taking into account
            chip count). This is usually true and is only useful in the context of WSOP 2021 Rule 96
            regarding when an all-in raise action does not trigger another round of betting.
        num_hands (int): The number of hands played so far.
        hand_phase (HandPhase): The current hand phase
        game_state (GameState): The current GameState
        hand_history (History): The history of the current hand.

    """

    def __init__(self, buyin: int, big_blind: int, small_blind: int, max_players=9):
        self.buyin = buyin
        self.big_blind = big_blind
        self.small_blind = small_blind
        self.max_players = max_players

        self.players: List[Player] = list(Player(i, self.buyin) for i in range(max_players))

        self.btn_loc = random.choice(self.players).player_id
        self.bb_loc = -1
        self.sb_loc = -1
        self.current_player = -1

        self.pots = []
        self._deck = None
        self.board = []
        self.hands = {}
        self.last_raise = 0
        self.raise_option = True

        self.num_hands = 0
        self.hand_phase = HandPhase.PREHAND
        self.game_state = GameState.RUNNING

        self._handstate_handler: Dict[HandPhase, Callable[[], Optional[Iterator[TexasHoldEm]]]] = {
            HandPhase.PREHAND: self._prehand,
            HandPhase.PREFLOP: lambda: self._betting_round(HandPhase.PREFLOP),
            HandPhase.FLOP: lambda: self._betting_round(HandPhase.FLOP),
            HandPhase.TURN: lambda: self._betting_round(HandPhase.TURN),
            HandPhase.RIVER: lambda: self._betting_round(HandPhase.RIVER),
            HandPhase.SETTLE: self._settle
        }

        self.hand_history: Optional[History] = None
        self._action = None, None
        self._hand_gen = None

    def _prehand(self):
        """
        Handles skips, not enough chips, rotation and posting of blinds,
        dealing cards and setup for preflop.

        """
        if self.hand_phase != HandPhase.PREHAND:
            raise ValueError("Not time for prehand!")

        # set statuses for players
        for player_id in self.player_iter(loc=0):
            self.players[player_id].last_pot = 0

            if self.players[player_id].chips == 0:
                self.players[player_id].state = PlayerState.SKIP
            else:
                self.players[player_id].state = PlayerState.TO_CALL

        active_players = list(self.in_pot_iter(self.btn_loc + 1))

        # stop if only 1 player
        if len(active_players) <= 1:
            self.game_state = GameState.STOPPED
            return

        # change btn loc (at least 2 players)
        self.btn_loc = active_players[0]
        self.sb_loc = active_players[1]

        # heads up edge case => sb = btn
        if len(active_players) == 2:
            self.sb_loc = self.btn_loc

        self.bb_loc = next(self.in_pot_iter(self.sb_loc + 1))

        # reset pots
        self.pots = [Pot()]

        # deal cards
        self._deck = Deck()
        self.hands = {}
        self.board = []

        for player_id in self.in_pot_iter(self.btn_loc + 1):
            self.hands[player_id] = self._deck.draw(num=2)

        # reset history
        self._action = None, None
        self.raise_option = True
        self.hand_history = History(
            prehand=PrehandHistory(
                btn_loc=self.btn_loc,
                big_blind=self.big_blind,
                small_blind=self.small_blind,
                player_chips={
                    i: self.players[i].chips
                    for i in range(self.max_players)
                },
                player_cards=self.hands
            )
        )

        # post blinds
        self._player_post(self.sb_loc, self.small_blind)
        self._player_post(self.bb_loc, self.big_blind)
        self.last_raise = 0

        # action to left of BB
        self.current_player = next(self.in_pot_iter(loc=self.bb_loc + 1))
        self.num_hands += 1

    def player_iter(self,
                    loc: int = None,
                    reverse: bool = False,
                    match_states: Iterable[PlayerState] = tuple(PlayerState),
                    filter_states: Iterable[PlayerState] = ()) -> Iterator[int]:
        """
        Iterates through all players starting at player_id and rotating in order
        of increasing player id.

        Arguments:
            loc (int, optional): The player_id to start at, default is :attr:`current_player`.
            reverse (bool): In reverse play order, default False
            match_states (Iterable[PlayerState]): Only include players with the given states
            filter_states (Iterable[PlayerState]): Exclude players with the given states
        Returns:
            Iterator[int]: An iterator over all player ids.

        """
        if loc is None:
            loc = self.current_player

        start, stop, step = loc, loc + self.max_players, 1
        if reverse:
            start, stop, step = stop, start, -step

        for i in range(start, stop, step):
            if self.players[i % self.max_players].state not in filter_states \
                    and self.players[i % self.max_players].state in match_states:
                yield i % self.max_players

    def in_pot_iter(self, loc: int = None, reverse: bool = False) -> Iterator[int]:
        """
        Iterates through all players with a stake in the pot (i.e. all players without
        states :obj:`~texasholdem.game.player_state.PlayerState.OUT` or
        :obj:`~texasholdem.game.player_state.PlayerState.SKIP`

        Arguments:
            loc (int, optional): The location to start at, defaults to current_player
            reverse (bool): In reverse play order, default False
        Returns:
            Iterator[int]: An iterator over players with a stake in the pot

        """
        if loc is None:
            loc = self.current_player
        yield from self.player_iter(loc=loc,
                                    reverse=reverse,
                                    filter_states=(PlayerState.OUT,
                                                   PlayerState.SKIP))

    def active_iter(self, loc: int = None, reverse: bool = False) -> Iterator[int]:
        """
        Iterates through all players that can take an action.
        i.e. players with states :obj:`~texasholdem.game.player_state.PlayerState.IN`
        or :obj:`~texasholdem.game.player_state.PlayerState.TO_CALL` (not including
        :obj:`~texasholdem.game.player_state.PlayerState.ALL_IN`).

        Arguments:
            loc (int, optional): The location to start at, defaults to current_player
            reverse (bool): In reverse play order, default False
        Returns:
            Iterator[int]: An iterator over active players who can take an action.

        """
        if loc is None:
            loc = self.current_player
        yield from self.player_iter(loc=loc,
                                    reverse=reverse,
                                    match_states=(PlayerState.TO_CALL,
                                                  PlayerState.IN))

    def _split_pot(self, pot_id: int, raised_level: int):
        """
        Splits the given pot at the given raised level, and adds players with
        excess to the new pot.

        Arguments:
            pot_id (int): The pot to split
            raised_level (int): The chip count to cut off at

        """
        pot = self._get_pot(pot_id)
        split_pot = pot.split_pot(raised_level)

        if not split_pot:
            return

        self.pots.insert(pot_id + 1, split_pot)

        for player_id in self.active_iter():
            if self.players[player_id].chips > self.chips_to_call(player_id):
                self.players[player_id].last_pot += 1

    def _player_post(self, player_id: int, amount: int):
        """
        Let a player post the given amount and sets the corresponding board state
        (i.e. makes other player states TO_CALL, sets ALL_IN). Also handles all
        pots (i.e. split pots).

        Arguments:
            player_id (int): The player_id of the player posting
            amount (int): The amount to post

        """
        amount = min(self.players[player_id].chips, amount)
        original_amount = amount
        last_pot = self.players[player_id].last_pot

        # if a player posts, they are in the pot
        if amount == self.players[player_id].chips:
            self.players[player_id].state = PlayerState.ALL_IN
        else:
            self.players[player_id].state = PlayerState.IN

        # call in previous pots
        for i in range(last_pot):
            amount = amount - self._get_pot(i).chips_to_call(player_id)
            self.pots[i].player_post(player_id, self.pots[i].chips_to_call(player_id))

        prev_raise_level = self.pots[last_pot].raised
        self._get_pot(last_pot).player_post(player_id, amount)

        last_raise = self.pots[last_pot].raised - prev_raise_level
        self.last_raise = max(last_raise, self.last_raise)

        # players previously in pot need to call in event of a raise
        if last_raise > 0:
            for pot_player_id in self._get_pot(last_pot).players_in_pot():
                if self._get_pot(last_pot).chips_to_call(pot_player_id) > 0 and \
                   self.players[pot_player_id].state == PlayerState.IN:
                    self.players[pot_player_id].state = PlayerState.TO_CALL

        # if a player is all_in in this pot, split a new one off
        if PlayerState.ALL_IN in (self.players[i].state
                                  for i in self._get_pot(last_pot).players_in_pot()):
            raised_level = min(self._get_pot(last_pot).get_player_amount(i)
                               for i in self._get_pot(last_pot).players_in_pot()
                               if self.players[i].state == PlayerState.ALL_IN)
            self._split_pot(last_pot, raised_level)

        self.players[player_id].chips = self.players[player_id].chips - original_amount

    def _get_pot(self, pot_id: int) -> Pot:
        """
        Arguments:
            pot_id (int): The id of the pot to get
        Returns:
            Pot: The pot with given id
        Raises:
            ValueError: If a pot with id pot_id does not exist.

        """
        if pot_id >= len(self.pots):
            raise ValueError(f"Pot with player_id {pot_id} does not exist.")

        return self.pots[pot_id]

    def _get_last_pot(self) -> Pot:
        """
        Returns:
            Pot: The current "active" pot

        """
        return self._get_pot(self._last_pot_id())

    def _last_pot_id(self) -> int:
        """
        Returns:
            int: The pot player_id of the last pot.

        """
        return len(self.pots) - 1

    def _is_hand_over(self) -> bool:
        """
        Returns:
            bool: True if no more actions can be taken by the remaining players.

        """
        count = 0
        for i in self.active_iter():
            if self.players[i].state == PlayerState.TO_CALL:
                return False

            if self.players[i].state == PlayerState.IN:
                count += 1

            if count > 1:
                return False
        return True

    def _settle(self):
        """
        Settles the current hand. If players are all-in, makes sure
        the board has 5 cards before settling.

        """
        if self.hand_phase != HandPhase.SETTLE:
            raise ValueError("Not time for Settle!")

        settle_history = SettleHistory(
            new_cards=[],
            pot_winners={}
        )
        self.hand_history[HandPhase.SETTLE] = settle_history

        self.current_player = next(self.in_pot_iter(loc=self.btn_loc + 1))

        for i, pot in enumerate(self.pots, 0):
            players_in_pot = list(pot.players_in_pot())
            # only player left in pot wins
            if len(players_in_pot) == 1:
                self.players[players_in_pot[0]].chips += pot.get_total_amount()
                settle_history.pot_winners[i] = (pot.get_total_amount(), -1, players_in_pot)
                continue

            # make sure there is 5 cards on the board
            if len(self.board) < 5:
                new_cards = self._deck.draw(num=5 - len(self.board))
                settle_history.new_cards.extend(new_cards)
                self.board.extend(new_cards)

            player_ranks = {}
            for player_id in players_in_pot:
                player_ranks[player_id] = evaluator.evaluate(self.hands[player_id], self.board)

            best_rank = min(player_ranks.values())
            winners = [player_id
                       for player_id, player_rank in player_ranks.items()
                       if player_rank == best_rank]

            settle_history.pot_winners[i] = (pot.get_total_amount(), best_rank, winners)

            win_amount = pot.get_total_amount() // len(winners)
            for player_id in winners:
                self.players[player_id].chips += win_amount

            # leftover chip goes to player left of the button WSOP Rule 73
            leftover = pot.get_total_amount() - (win_amount * len(winners))
            if leftover:
                for j in self.in_pot_iter(loc=self.btn_loc + 1):
                    if j in winners:
                        self.players[j].chips += leftover
                        break

    def chips_to_call(self, player_id: int) -> int:
        """
        Arguments:
            player_id (int): The player id
        Returns:
            int: The amount of chips the player needs to call in all pots to play the hand.

        """
        return sum(self._get_pot(i).chips_to_call(player_id)
                   for i in range(self.players[player_id].last_pot + 1))

    def player_bet_amount(self, player_id: int) -> int:
        """
        Arguments:
            player_id (int): The player player_id
        Returns:
            int: The amount of chips the player bet this round across all pots.

        """
        return sum(self._get_pot(i).get_player_amount(player_id) for i in range(len(self.pots)))

    def chips_at_stake(self, player_id: int) -> int:
        """
        Arguments:
            player_id (int) - The player player_id
        Returns:
            int: The amount of chips the player is eligible to win

        """
        return sum(self._get_pot(i).get_total_amount()
                   for i in range(len(self.pots))
                   if player_id in self._get_pot(i).players_in_pot())

    @versionadded(version="0.6.0")
    def total_to_value(self, total: Optional[int], player_id: int) -> Optional[int]:
        """
        Translates a raise phrase "raise *to* total" to the phrase
        "raise value".

        Arguments:
            total (int, optional): A total amount to raise
            player_id (int): The player who's doing the raising
        Returns:
            Optional[int]: A translation to an amount the player is raising *by*
        """
        if not total:
            return None
        return total - self.chips_to_call(player_id) - self.player_bet_amount(player_id)

    @versionadded(version="0.6.0")
    def value_to_total(self, value: Optional[int], player_id: int) -> Optional[int]:
        """
        Translates a raise phrase "raise value" to the phrase
        "raise *to* total".

        Arguments:
            value (int, optional): An amount to raise past the previous bet
            player_id (int): The player who's doing the raising
        Returns:
            Optional[int]: A translation to an amount the player is raising *to*
        """
        if not value:
            return None
        return value + self.chips_to_call(player_id) + self.player_bet_amount(player_id)

    @versionadded(version="0.6.0")
    def min_raise(self):
        """
        Returns:
            The minimum amount a player can raise by.
        """
        return max(self.big_blind, self.last_raise)

    @check_raise(ValueError)
    @versionchanged(reason=
                    "The :code:`value` has been renamed to :code:`total` and will "
                    "be redefined in 1.0.0. Currently, :code:`value` and :code:`total` "
                    "mean to raise *to* the amount given. In 1.0.0, :code:`value` will "
                    "mean to raise an amount more than the current bet amount.",
                    version="0.6.0")
    def validate_move(self,
                      player_id: Optional[int] = None,
                      action: Optional[ActionType] = None,
                      value: Optional[int] = None,
                      total: Optional[int] = None,
                      throws: bool = False):
        # pylint: disable=unused-argument,too-many-arguments,
        # pylint: disable=too-many-branches,too-many-return-statements
        """
        Validate the potentially invalid action for the given player.

        .. note::
            :code:`value` and :code:`total` are mutually exclusive.

        Arguments:
            player_id (int, optional): The player to take action. Default current_player.
            action (int, optional): The ActionType to take
            value (int, optional): For a raise action, how much to raise *to*.
            total (int, optional): For a raise action, how much to raise *to*.
            throws (bool): If true, will throw an Exception instead if
                the move is invalid. Default False.
        Returns:
            bool: True if the move is valid, False otherwise.
        Raises:
            ValueError: If move is invalid and throws is True

        """
        if player_id is None:
            player_id = self.current_player

        if total and value:
            raise ValueError("Got arguments for both total and value. Expected only one.")

        if value:
            warnings.warn(
                "The :code:`value` has been renamed to :code:`total` and will "
                "be redefined in 1.0.0. Currently, :code:`value` and :code:`total` "
                "mean to raise *to* the amount given. In 1.0.0, :code:`value` will "
                "mean to raise an amount more than the current bet amount.",
                DeprecationWarning)
            total = value

        # ALL_IN should be translated
        new_action, new_total = action, total
        if new_action == ActionType.ALL_IN:
            new_action, new_total = self._translate_allin(new_action, new_total)

        player_amount = self.player_bet_amount(player_id)
        chips_to_call = self.chips_to_call(player_id)

        if not action:
            return False, "Action is None."

        if self.current_player != player_id:
            return False, f"Player {player_id} is not the current " \
                          f"player (Current player {self.current_player})"

        if new_action == ActionType.CALL and \
                self.players[player_id].state != PlayerState.TO_CALL:
            return False, f"Player {player_id} has state " \
                          f"{self.players[player_id].state.name} cannot CALL"

        if new_action == ActionType.CHECK and \
                self.players[player_id].state != PlayerState.IN:
            return False, \
                   f"Player {player_id} has state {self.players[player_id].state.name} cannot CHECK"

        if new_action == ActionType.RAISE:
            if new_total is None or new_total == 0:
                return False, "Expected value to not be None or 0 for action RAISE."

            if not self.raise_option:
                return False, "Betting round is over at this point, can only CALL or FOLD."

            if (self.total_to_value(new_total, player_id) < self.min_raise()
               and new_total < player_amount + self.players[player_id].chips):
                return False, f"Cannot raise {self.total_to_value(new_total, player_id)}, " \
                              f"less than the min raise {self.min_raise()} and player " \
                              f"{player_id} is not going all-in."
            if player_amount + self.players[player_id].chips < new_total:
                return False, f"Cannot raise {new_total}, more than the number of chips " \
                              f"available {player_amount + self.players[player_id].chips}"
            if new_total < chips_to_call:
                return False, f"Expected raise value {new_total} to be more " \
                              f"than the chips to call {chips_to_call}"
        return True, ""

    def _take_action(self,
                     action: ActionType,
                     total: Optional[int] = None):
        """
        Execute the action for the current player. Assumes the move is valid.

        Arguments:
            action (ActionType): The ActionType to take
            total (int, optional): In the case of raise, how much to raise *to*

        """
        # ALL_IN should be translated
        if action == ActionType.ALL_IN:
            action, total = self._translate_allin(action, total=total)

        player_amount = self.player_bet_amount(self.current_player)
        chips_to_call = self.chips_to_call(self.current_player)

        # Execute move
        if action == ActionType.CALL:
            self._player_post(self.current_player, chips_to_call)
        elif action == ActionType.CHECK:
            pass
        elif action == ActionType.RAISE:
            self._player_post(self.current_player, total - player_amount)
        elif action == ActionType.FOLD:
            self.players[self.current_player].state = PlayerState.OUT
            for i in range(self.players[self.current_player].last_pot + 1):
                self.pots[i].remove_player(self.current_player)

    def _translate_allin(self,
                         action: ActionType,
                         total: int = None) -> Tuple[ActionType, Optional[int]]:
        """
        Translates an all-in action to the appropriate action,
        either call or raise.

        """
        if action != ActionType.ALL_IN:
            return action, total

        if self.players[self.current_player].chips <= self.chips_to_call(self.current_player):
            return ActionType.CALL, None

        return ActionType.RAISE, \
            self.player_bet_amount(self.current_player) + self.players[self.current_player].chips

    def _previous_all_in_sum(self) -> int:
        """
        Used for WSOP Rule 96

        Returns:
            int: The sum of the raise values of the most recent all-in raise players.
        """
        raised_sum = 0

        for action in reversed(self.hand_history[self.hand_phase].actions):
            if self.players[action.player_id].state == PlayerState.ALL_IN \
                    and action.action_type == ActionType.RAISE:
                raised_sum += action.value
            elif self.players[action.player_id].state == PlayerState.IN:
                break

        return raised_sum

    def _betting_round(self, hand_phase: HandPhase) -> Iterator[TexasHoldEm]:
        """
        Core round of the poker game. Executes actions from each active player
        until everyone "checks"

        Arguments:
            hand_phase (HandPhase): Which betting round phase to execute
        Returns:
            Iterator[TexasHoldEm]: An iterator over the intermediary game states right when
                a player is expected to take a turn.
        Raises:
            ValueError: If self.hand_state is not a valid betting round

        """
        if hand_phase not in (HandPhase.PREFLOP, HandPhase.FLOP, HandPhase.TURN, HandPhase.RIVER):
            raise ValueError("Not valid betting round!")

        if hand_phase != self.hand_phase:
            raise ValueError(f"Hand phase mismatch: expected {self.hand_phase}, got {hand_phase}")

        # add new cards to the board
        new_cards = self._deck.draw(num=hand_phase.new_cards())
        self.hand_history[hand_phase] = BettingRoundHistory(
            new_cards=new_cards,
            actions=[]
        )
        self.board.extend(new_cards)

        # reset last raise
        first_pot = self._last_pot_id()
        self.last_raise = 0
        self.raise_option = True

        # player to the left of the button starts
        self.current_player = self.btn_loc + 1
        if hand_phase == HandPhase.PREFLOP:
            self.current_player = self.bb_loc + 1

        player_queue = deque(self.active_iter(self.current_player))

        while not self._is_hand_over():
            # WSOP 2021 Rule 96
            # if no more active players that can raise continue with the players to call
            # while disabling the raise availability.
            if not player_queue:
                player_queue = deque(self.player_iter(loc=self.current_player + 1,
                                                      match_states=(PlayerState.TO_CALL,)))
                if not player_queue:
                    break

                self.raise_option = False

            # book keeping
            prev_raised = self.last_raise

            # set the current player and yield
            self.current_player = player_queue.popleft()
            yield self

            action, total = self._translate_allin(*self._action)
            value = self.total_to_value(total=total,
                                        player_id=self.current_player)
            self.validate_move(action=action, total=total, throws=True)

            betting_history = self.hand_history[self.hand_phase]
            betting_history.actions.append(PlayerAction(
                player_id=self.current_player,
                action_type=action,
                total=total,
                value=value)
            )

            self._take_action(action, total=total)

            # On raise, everyone eligible gets to take another action
            if action == ActionType.RAISE:
                # WSOP 2021 Rule 96
                # An all-in raise less than the previous raise shall not reopen
                # the bidding unless two or more such all-in raises total greater
                # than or equal to the previous raise.
                raise_sum = self._previous_all_in_sum()
                if value < prev_raised:
                    if raise_sum < prev_raised:
                        continue
                    # Exception for rule 96, set this
                    self.last_raise = raise_sum

                # reset the round (i.e. as if the betting round started here)
                player_queue = deque(self.active_iter(self.current_player))

                # Throwaway current player
                # Edge case: active_iter already excludes ALL_IN
                if self.players[self.current_player].state != PlayerState.ALL_IN:
                    player_queue.popleft()

        # consolidate betting to all pots in this betting round
        for i in range(first_pot, len(self.pots)):
            self._get_pot(i).collect_bets()

    def get_hand(self, player_id) -> List[Card]:
        """
        Arguments:
            player_id (int): The player id
        Returns:
            List[Card]: A two element list of the hand of the given player,
                if the player has not been dealt a hand, returns an empty list

        """
        return self.hands.get(player_id, [])

    def start_hand(self):
        """
        Starts a new hand. Handles :obj:`~texasholdem.game.player_state.PlayerState.SKIP`,
        not enough chips, resetting pots, rotation and posting of blinds, dealing cards, etc.

        Raises:
            ValueError: If a hand already in progress.

        """
        if self.is_hand_running():
            raise ValueError('In the middle of a hand!')

        self.hand_phase = HandPhase.PREHAND
        self._handstate_handler[self.hand_phase]()

        if self.game_state == GameState.STOPPED:
            return

        self.hand_phase = self.hand_phase.next_phase()
        self._hand_gen = self._hand_iter()

        try:
            next(self._hand_gen)
        except StopIteration:
            pass

    @versionchanged(reason=
                    "The :code:`value` has been renamed to :code:`total` and will "
                    "be redefined in 1.0.0. Currently, :code:`value` and :code:`total` "
                    "mean to raise *to* the amount given. In 1.0.0, :code:`value` will "
                    "mean to raise an amount more than the current bet amount.",
                    version="0.6.0")
    def take_action(self,
                    action_type: ActionType,
                    value: Optional[int] = None,
                    total: Optional[int] = None):
        """
        The current player takes the specified action.

        .. note::
            :code:`value` and :code:`total` are mutually exclusive.

        Arguments:
            action_type (ActionType): The action type
            value (int, optional): For a raise action, how much to raise *to*.
            total (int, optional): For a raise action, how much to raise *to*.
        Raises:
            ValueError: If no hand is running or if the move is invalid.

        """
        if not self.is_hand_running():
            raise ValueError("No hand is running")

        if total and value:
            raise ValueError("Got arguments for both total and value. Expected only one.")

        if value:
            warnings.warn(
                "The :code:`value` has been renamed to :code:`total` and will "
                "be redefined in 1.0.0. Currently, :code:`value` and :code:`total` "
                "mean to raise *to* the amount given. In 1.0.0, :code:`value` will "
                "mean to raise an amount more than the current bet amount.",
                DeprecationWarning)
            total = value

        self.validate_move(action=action_type, total=total, throws=True)
        self._action = (action_type, total)

        try:
            next(self._hand_gen)
        except StopIteration:
            pass

    def _hand_iter(self) -> Iterator[TexasHoldEm]:
        """
        Returns:
            Iterator[TexasHoldEm]: A generator over every intermediate game state.
                i.e. right before every action.
        Raises:
            ValueError: If phase != PREFLOP

        """
        if self.hand_phase != HandPhase.PREFLOP:
            raise ValueError("Cannot iterate over hand: not time for PREFLOP")

        while self.is_hand_running():
            if self._is_hand_over():
                self.hand_phase = HandPhase.SETTLE

            round_iter: Optional[Iterator[TexasHoldEm]] = self._handstate_handler[self.hand_phase]()
            if round_iter:
                yield from round_iter

            self.hand_phase = self.hand_phase.next_phase()

    def is_hand_running(self) -> bool:
        """
        Returns:
            bool: True if there is a hand running, false o/w

        """
        return self.hand_phase != HandPhase.PREHAND

    def is_game_running(self) -> bool:
        """
        Returns:
            bool: True if the game is running, false o/w

        """
        return self.game_state == GameState.RUNNING

    def export_history(self, path: Union[str, os.PathLike] = "./texas.pgn") -> os.PathLike:
        """
        Exports the hand history to a human-readable file. If a directory is given,
        finds a name of the form :code:`texas(n).pgn` to export to. PGN files can consequently
        be imported with :meth:`import_history()`.

        Arguments:
            path (str | os.PathLike]): The directory or file to export the history to,
                defaults to the current working directory at the file `./texas.pgn`
        Returns:
            os.PathLike: The path to the history file

        """
        return self.hand_history.export_history(path)

    @staticmethod
    def import_history(path: Union[str, os.PathLike]) -> Iterator[TexasHoldEm]:
        """
        Import a PGN file i.e. as exported from :meth:`export_history()`. Returns an
        iterator over game states as specified from the history.

        Arguments:
            path (str | os.PathLike): The PGN file to import from
        Returns:
            Iterator[TexasHoldEm]: An iterator over game states such that
                the next hand will play exactly like from the history.
        Raises:
            HistoryImportError: If the file given does not exist or if the file is invalid

        """
        return TexasHoldEm._import_history(History.import_history(path))

    @staticmethod
    def _import_history(history: History) -> Iterator[TexasHoldEm]:
        """
        Arguments:
            history (History): The History file to import from
        Returns:
            Iterator[TexasHoldEm]: An iterator over game states such that
                the next hand will play exactly like from the history.
        Raises:
            HistoryImportError: If there was an error running the history.

        """
        # pylint: disable=protected-access
        num_players = len(history.prehand.player_chips)
        game = TexasHoldEm(buyin=1,
                           big_blind=history.prehand.big_blind,
                           small_blind=history.prehand.small_blind,
                           max_players=num_players)

        # button placed right before 0
        game.btn_loc = num_players - 1

        # read chips
        for i in game.player_iter(0):
            game.players[i].chips = history.prehand.player_chips[i]

        # stack deck
        deck = Deck()
        deck.cards = list(history.settle.new_cards)

        # player actions in a stack
        player_actions: List[Tuple[int, ActionType, Optional[int]]] = []
        for bet_round in (history.river,
                          history.turn,
                          history.flop,
                          history.preflop):
            if bet_round:
                deck.cards = bet_round.new_cards + deck.cards
                for action in reversed(bet_round.actions):
                    player_actions.insert(0, (action.player_id, action.action_type, action.total))

        # start hand (deck will deal)
        game.start_hand()

        # give players old cards
        for i in game.player_iter():
            game.hands[i] = history.prehand.player_cards[i]

        # swap decks
        game._deck = deck

        while game.is_hand_running():
            yield game

            try:
                player_id, action, total = player_actions.pop(0)
            except IndexError as err:
                raise HistoryImportError("Expected more actions than "
                                         "given in the history file.") from err

            if player_id != game.current_player:
                raise HistoryImportError(f"Error replaying history: action player {player_id} "
                                         f"is not current player {game.current_player}")

            game.take_action(action, total=total)
        yield game



"""
Basic agents are included in this module:
    - :func:`call_agent`
    - :func:`random_agent`

"""

from typing import Tuple

import random

from texasholdem.game.game import TexasHoldEm
from texasholdem.game.action_type import ActionType
from texasholdem.game.player_state import PlayerState


def call_agent(game: TexasHoldEm) -> Tuple[ActionType, None]:
    """
    A player that calls if another player raised or checks.

    Arguments:
        game (TexasHoldEm): The TexasHoldEm game
    Returns:
        Tuple[ActionType, None]: CALL if someone raised, else CHECK

    """
    player = game.players[game.current_player]
    if player.state == PlayerState.TO_CALL:
        return ActionType.CALL, None
    return ActionType.CHECK, None


def random_agent(game: TexasHoldEm, no_fold: bool = False) -> Tuple[ActionType, int]:
    """
    A uniformly random player

        - If someone raised, CALL, FOLD, or RAISE with uniform probability
        - Else, CHECK, (FOLD if no_fold=False), RAISE with uniform probability
        - If RAISE, the value will be uniformly random in [min_raise, # of chips]

    Arguments:
        game (TexasHoldEm): The TexasHoldEm game
        no_fold (bool): Removes the possibility of folding if no one raised, default False.
    Returns:
        Tuple[ActionType, int]: Returns a uniformly random action from the
            available moves.

    """
    bet_amount = game.player_bet_amount(game.current_player)
    chips = game.players[game.current_player].chips
    min_raise = game.value_to_total(game.min_raise(), game.current_player)
    max_raise = bet_amount + chips

    possible = list(ActionType)
    possible.remove(ActionType.ALL_IN)

    # A player did not raise
    if game.players[game.current_player].state == PlayerState.IN:
        possible.remove(ActionType.CALL)
        if no_fold:
            possible.remove(ActionType.FOLD)

    # A player raised
    if game.players[game.current_player].state == PlayerState.TO_CALL:
        possible.remove(ActionType.CHECK)

    # not enough chips to raise
    if max_raise < min_raise:
        possible.remove(ActionType.RAISE)

    action_type, total = random.choice(possible), None
    if action_type == ActionType.RAISE:
        total = random.randint(min_raise, max_raise)

    return action_type, total



import abc
import os
import logging
from typing import Optional, Iterable, Tuple, Union

from deprecated.sphinx import versionadded

from texasholdem.game.action_type import ActionType
from texasholdem.game.game import TexasHoldEm


logger = logging.getLogger(__name__)


class AbstractGUI(abc.ABC):
    """
    This class provides a recommended outline of the methods that every TexasHoldEm GUI
    should implement. It also comes with a few convenience methods for implementations,
    including :meth:`run_step` which runs a complete step of the game (display state,
    take input, show state, etc.) and :meth:`replay_history` which allows the user
    to step through the given history.

    Arguments:
        game (TexasHoldEm, optional): The game object to attach to, all methods will
            default to this game. (Not necessary if only showing the history)
        visible_players (Iterable[int], optional): The players whose cards should be
            displayed whenever the :meth:`display_state` method is called, defaults to every
            player.
        enable_animation (bool): If set to True, will play animations, default True.
        no_wait (bool): If set to True, disables waiting mechanisms and will not block.
    Attributes:
        game (TexasHoldEm, optional): The game object to attach to, all methods will
            default to this game. (Not necessary if only showing the history)
        visible_players (list[int], optional): The players whose cards should be displayed
            whenever the :meth:`display_state` method is called, defaults to every player.
        enable_animation (bool): If set to True, will play animations, default True.
        no_wait (bool): If set to True, disables waiting mechanisms and will not block.

    """

    def __init__(self,
                 game: Optional[TexasHoldEm] = None,
                 visible_players: Optional[Iterable[int]] = None,
                 enable_animation: bool = True,
                 no_wait: bool = False):
        self.game = game
        self.visible_players = visible_players
        self.no_wait = no_wait
        self.enable_animation = enable_animation

        # All players visible by default
        if visible_players is None and game:
            self.set_visible_players(range(self.game.max_players))

    @versionadded(version="0.7.0")
    def set_visible_players(self, visible_players: Iterable[int]):
        """
        Make the given players' cards visible.

        Arguments:
            visible_players (Iterable[int]): The players whose cards should be visible when the
                :meth:`display_state` method is called.
        Raises:
            ValueError: If any of the player ids given are invalid.

        """
        if not visible_players:
            self.visible_players = visible_players
            return

        sorted_players = sorted(list(visible_players))
        if len(sorted_players) > self.game.max_players:
            raise ValueError("Expected length of visible players to be <= number of players. "
                             f"Expected <= {self.game.max_players}, Got {len(sorted_players)}.")

        for player_id, criteria in ((sorted_players[0], lambda i: i < 0),
                                    (sorted_players[-1], lambda i: i > self.game.max_players - 1)):
            if criteria(player_id):
                raise ValueError(f"Unexpected player id {player_id}")

        self.visible_players = sorted_players

    @versionadded(version="0.7.0")
    def prompt_input(self):
        """
        Prompt the user for input.

        """

    @versionadded(version="0.7.0")
    def accept_input(self) -> Tuple[ActionType, Optional[int]]:
        """
        Receive input from the user and translate the given input to the canonical
        (ActionType, int) tuple form.

        Implementations should only focus on this direct translation and not worry about
        validating the move with respect to the game. Implementations should also raise a
        ValueError if the given user input is malformed.

        Returns:
            Tuple[ActionType, Optional[int]]: The action
        Raises:
            ValueError: If the given input could not be parsed

        """
        raise NotImplementedError()

    @versionadded(version="0.7.0")
    def hide(self):
        """
        Hide the GUI.

        """
        raise NotImplementedError()

    @versionadded(version="0.7.0")
    def refresh(self):
        """
        Refresh the GUI.

        """

    @versionadded(version="0.7.0")
    def wait_until_prompted(self):
        """
        Wait until the user gives the appropriate signal.

        """

    @versionadded(version="0.7.0")
    def display_state(self):
        """
        Display the state of the game.

        """
        raise NotImplementedError()

    @versionadded(version="0.7.0")
    def display_error(self, error: str):
        """
        Display any potential errors from users (malformed input, invalid action, etc.)

        Arguments:
            error (str): The error message

        """

    @versionadded(version="0.7.0")
    def display_action(self):
        """
        Display the most recent action

        """

    @versionadded(version="0.7.0")
    def display_win(self):
        """
        Display the winners of the hand, (and everyone's cards that didn't fold)

        """
        raise NotImplementedError()

    @versionadded(version="0.7.0")
    def run_step(self):
        """
        Runs a complete GUI step of the hand:

            - Display the game state
            - Prompt for action from the user until valid
            - Take the action and display it
            - Display the winners if the hand ended

        This is included as a convenience method built from the other necessary
        abstract methods.

        """
        if not self.game.is_hand_running():
            return

        self.display_state()

        # Prompt for action input until valid
        while True:
            try:
                self.prompt_input()
                action, total = self.accept_input()
                self.game.validate_move(action=action,
                                        total=total,
                                        throws=True)
                break
            except ValueError as err:
                logger.warning("Caught error: %s.", str(err), exc_info=err)
                self.display_error(str(err))

        # Take the action in the game
        self.game.take_action(action, total=total)

        # Announce the move
        self.display_action()

        # display game after running action
        self.display_state()

        # Display the winners if the hand ended
        if not self.game.is_hand_running():
            self.display_win()

    @versionadded(version="0.7.0")
    def replay_history(self, path: Union[str, os.PathLike]):
        """
        Replays the given history, going forward when prompted.

        This is included as a convenience method built from the other necessary
        abstract methods.

        """
        old_game = self.game
        old_visible_players = self.visible_players

        for state in TexasHoldEm.import_history(path):
            self.game = state
            self.set_visible_players(range(self.game.max_players))
            self.refresh()
            self.display_action()
            self.display_state()
            self.wait_until_prompted()

        self.game = old_game
        self.visible_players = old_visible_players



# pylint: disable=invalid-name
from __future__ import annotations
import enum
import logging
import math
import platform
import shutil
import re
import sys
from typing import Iterable, Optional, Union, Tuple
import curses
from collections import namedtuple, deque
import signal
from importlib.metadata import version

from deprecated.sphinx import deprecated

from texasholdem.util.errors import Ignore
from texasholdem.util.functions import preflight, handle, raise_if
from texasholdem.card import card
from texasholdem.game.game import TexasHoldEm
from texasholdem.game.action_type import ActionType
from texasholdem.game.hand_phase import HandPhase
from texasholdem.game.player_state import PlayerState
from texasholdem.gui.abstract_gui import AbstractGUI


# Windows Compatibility
_OS = platform.system()
_IS_WINDOWS = _OS == 'Windows'

if _IS_WINDOWS:
    curses.resizeterm = curses.resize_term


logger = logging.getLogger(__name__)


_BlockDim = namedtuple("_BlockDim", ['rows', 'cols'])


class _Ellipse:
    """
    Represents an ellipse, contains methods to
        - Get a :meth:`point_yx` from radians
        - Get the :meth:`derivative` from radians
        - Get the :meth:`char_at` radians which describes the derivative

    Arguments:
        major (float): The semi-major axis length
        minor (float): The semi-minor axis length
        center (Tuple[float, float]): The center of the ellipse

    """

    def __init__(self,
                 major: float = None,
                 minor: float = None,
                 center: Tuple[float, float] = (0, 0)):
        self.major = major
        self.minor = minor
        self.center = center

    def point_yx(self, rads: float) -> Tuple[float, float]:
        """
        Arguments:
            rads (float): The radians
        Returns:
            Tuple[float, float]: The point y, x

        """
        return (self.minor * math.sin(rads) + self.center[1],
                self.major * math.cos(rads) + self.center[0])

    def derivative(self, rads: float) -> float:
        """
        Arguments:
            rads (float): The radians
        Returns:
            float: The derivative dy/dx

        """
        return (rads * self.minor * math.cos(rads)) / (rads * self.major * -math.sin(rads))

    def char_at(self, rads: float) -> str:
        """
        Arguments:
            rads (float): The radians
        Returns:
            str: The character that describes the derivative at the point

        """
        derivative = self.derivative(rads=rads)
        if derivative < -0.5:
            return "|"
        if derivative < -0.25:
            return "/"
        if derivative < 0.25:
            if math.sin(rads) >= 0:
                return "_"
            return "‾"
        if derivative < 0.5:
            return "\\"
        return "|"


class _Align(enum.Enum):
    """
    Enum that represents the alignment in box top/middle/bottom

    """
    TOP = enum.auto()
    MIDDLE = enum.auto()
    BOTTOM = enum.auto()


class _Justify(enum.Enum):
    """
    Enum that represents how a text is justified in a box
    left, center, right.

    """
    LEFT = enum.auto()
    CENTER = enum.auto()
    RIGHT = enum.auto()


class _Block:
    """
    Core class of the Text GUI system. Wraps around the curses._CursesWindow object
    to provide helper functions that makes working with text, centering, resizing,
    erasing, and working with nested windows easier.

    Arguments:
        name (str): The name of the block element.
        window (curses._CursesWindow): The window object for this block (usually given
            from another :class:`_Block` with the :meth:`new_block` call (which
            calls :code:`curses.newwin`).
    Attributes:
        name (str): The name of the block element.
        window (curses._CursesWindow): The window object for this block (usually given
            from another :class:`_Block` or :class:`_CursesHelper` with the :meth:`new_block`
            call (which calls :code:`curses.newwin`).
        blocks (dict[str, _Block]): a dictionary of child blocks.
        content (list, dict): The last *args, and **kwargs passed in to the :meth:`add_content`
            method. Used to refresh and save state.
        content_stack (deque): A stack of the previous contents saved with :meth:`stash_state`

    """

    def __init__(self,
                 name: str,
                 window: curses._CursesWindow = None):
        # pylint: disable=no-member
        self.name: str = name
        self.blocks: dict[str, _Block] = {}
        self.window = window
        self.content_stack = deque(maxlen=10)
        self.content = None

    def _set_content_call(self, *args, **kwargs):
        self.content = (args, kwargs)

    @staticmethod
    def _pad(obj: Union[list[str], str],
             pad_obj: Union[list[str], str],
             padding_len: int,
             min_padding: int,
             align: _Align) -> Union[list[str], str]:
        # pylint: disable=too-many-arguments
        """
        Helper function to pad a list or string

        """
        if align == _Align.BOTTOM:
            before_padding = pad_obj * max(padding_len, min_padding)
            after_padding = pad_obj * min_padding
        elif align == _Align.MIDDLE:
            before_padding = after_padding = pad_obj * max(padding_len // 2, min_padding)
        else:
            after_padding = pad_obj * max(padding_len, min_padding)
            before_padding = pad_obj * min_padding

        return before_padding + obj + after_padding

    @handle(handler=lambda exc: logger.debug(str(exc), exc_info=exc),
            exc_type=curses.error)
    @preflight(prerun=lambda self, *args, **kwargs: self._set_content_call(*args, **kwargs))
    def add_content(self,
                    content: list[str],
                    align: _Align = _Align.MIDDLE,
                    justify: _Justify = _Justify.CENTER,
                    border: bool = False,
                    wrap_line: bool = False):
        # pylint: disable=too-many-arguments
        """
        Add the given string list to the block. Each element is placed on a new line.
        Pass in parameters to modify the alignment, justification, border, etc.

        Arguments:
            content (list[str]): The content to add to the block, each element is a
                new line.
            align (_Align): The alignment (top to bottom) for the content.
            justify (_Justify): The justification (left to right) for the content.
            wrap_line (bool): Set to True to split lines that would extend past the box
                boundary. Set to False to replace any overflow with '...'. Defaults to False
            border (bool): Set to True to add a border, default False.

        """
        self.window.erase()
        rows, cols = self.window.getmaxyx()
        border_int = 1 if border else 0

        if wrap_line:
            for i, text in enumerate(content):
                if len(text) >= cols:
                    end = cols-len(_DOTS)-1
                    content[i] = text[:end]
                    content.insert(i+1, text[end:])

        # align top/middle/bottom
        content = self._pad(obj=content,
                            pad_obj=[""],
                            padding_len=rows - len(content) - 1,
                            min_padding=border_int,
                            align=align)

        # align left/center/right
        for i, text in enumerate(content):
            if len(text) >= cols:
                continue

            if justify == _Justify.RIGHT:
                align = _Align.BOTTOM
            elif justify == _Justify.CENTER:
                align = _Align.MIDDLE
            else:
                align = _Align.TOP

            content[i] = self._pad(obj=content[i],
                                   pad_obj=" ",
                                   padding_len=cols - len(text) - 1,
                                   min_padding=border_int,
                                   align=align)

        # right out, masking with dots if too long
        for i, text in enumerate(content):
            if len(text) >= cols:
                text = text[:(cols-1-len(_DOTS))] + _DOTS
            text += '\n' if i != rows - 1 else ''
            self.window.addstr(text)

        # add border
        if border:
            self.window.border(*_BLOCK_BORDER)

    def stash_state(self):
        """
        Saves the current content call onto the :attr:`content_stack` and
        calls :meth:`stash_state` on any child blocks.

        """
        if self.content:
            self.content_stack.appendleft(self.content)
        for block in self.blocks.values():
            block.stash_state()

    def pop_state(self):
        """
        Pops from the :attr:`content_stack` and restores the content call and calls
        :meth:`pop_state` on any child blocks.

        .. note::
            Will be a noop if :attr:`content_stack` is empty

        """
        if self.content_stack:
            args, kwargs = self.content_stack.popleft()
            self.add_content(*args, **kwargs)
        for block in self.blocks.values():
            block.pop_state()

    @handle(handler=lambda exc: logger.debug(str(exc), exc_info=exc),
            exc_type=curses.error)
    def new_block(self,
                  name: str,
                  nlines: int,
                  ncols: int,
                  begin_y: int = 0,
                  begin_x: int = 0) -> _Block:
        # pylint: disable=too-many-arguments
        """
        Creates and returns a new block (that wraps around curses._CursesWindow). Note
        that this method is smart and so if the given block already exists, it will
        resize and move the block with the given arguments.

        Arguments:
            name (str): The name to give to the child block
            nlines (int): The number of rows to give the new block
            ncols (int): The number of columns to give to the new block
            begin_y (int): The topleft y coordinate of the block
            begin_x (int): The topleft x coordinate of the block
        Returns:
            _Block: The newly created block (or the existing block

        """

        if name in self.blocks:
            self.blocks[name].window.resize(nlines, ncols)
            self.blocks[name].window.mvwin(*self.bound_coords(begin_y, begin_x))
            return self.blocks[name]

        block = _Block(name=name,
                       window=curses.newwin(nlines, ncols,
                                            *self.bound_coords(begin_y, begin_x)))
        self.blocks[name] = block
        return block

    def get_block(self, name: str) -> Optional[_Block]:
        """
        Get the child block by name (also searches sub-children)

        Arguments:
            name (str): The block name to get
        Returns:
            Optional[_Block]: The _Block by name or None

        """
        if name in self.blocks:
            return self.blocks[name]
        for block in self.blocks.values():
            child = block.get_block(name)
            if child:
                return child
        return None

    def erase(self):
        """
        Erases the window and unsets the :attr:`content` attributes.

        .. note::
            You should call this method :code:`block.erase()` instead of
            :code:`block.window.erase()` as it does not erase the content from
            the stack.

        """
        self.content = None
        self.window.erase()

    def refresh(self):
        """
        Refreshes the block window and any child blocks.

        """
        self.window.refresh()
        for block in self.blocks.values():
            block.refresh()

    def bound_coords(self, y: int, x: int) -> Tuple[int, int]:
        """
        Ensures the given y, x will lay in the window.

        Arguments:
            y (int): The y coordinate
            x (int): The x coordinate
        Returns:
            Tuple[int, int]: The safe bounded coordinates

        """
        max_y, max_x = self.window.getmaxyx()
        y_start, x_start = self.window.getbegyx()
        return (min(max(y_start, y), y_start + max_y),
                min(max(x_start, x), x_start + max_x))


# STRING CONSTANTS
_PROMPT = '$ '
_BLOCK_BORDER = ('|', '|', '-', '-', '+', '+', '+', '+')
_DOTS = '...'

# BLOCK DIMENSIONS
_PLAYER_BLOCK_SIZE = _BlockDim(rows=7, cols=20)
_PLAYER_BET_BLOCK_SIZE = _BlockDim(rows=1, cols=10)
_BOARD_BLOCK_SIZE = _BlockDim(rows=5, cols=50)
_PROMPT_BLOCK_SIZE = _BlockDim(rows=2, cols=35)
_ERROR_BLOCK_SIZE = _BlockDim(rows=1, cols=80)
_HISTORY_BLOCK_SIZE = _BlockDim(rows=-1, cols=28)
_ACTION_BLOCK_SIZE = _BlockDim(rows=1, cols=2)
_VERSION_BLOCK_SIZE = _BlockDim(rows=1, cols=25)

# OFFSETS & SIZE MULTIPLIERS
_HISTORY_BLOCK_SIZE_FACTOR = 0.95
_TABLE_ELLIPSE_OFFSET = (-15, -2)
_PLAYER_ELLIPSE_SIZE_FACTOR = 0.72
_TABLE_ELLIPSE_SIZE_FACTOR = 0.5
_PLAYER_BET_ELLIPSE_SIZE_FACTOR = 0.35
_TABLE_STEPS_RESOLUTION = 400

# KEY STROKES
_BACKSPACE = 127 if not _IS_WINDOWS else 8
_NEWLINE = 10
_RESIZE = -1
_CTRL_C = 3         # Windows Only

# ANIMATION TIMING
_ACTION_STEPS = 10
_ACTION_SLEEP_MS = 20 if not _IS_WINDOWS else 10


class TextGUI(AbstractGUI):
    """
    Text-based GUI. Play Texas Hold 'Em on the command line.

    Arguments:
        game (TexasHoldEm, optional): The game object to attach to, all methods will
            default to this game. (Not necessary if only showing the history)
        visible_players (Iterable[int], optional): The players whose cards should be
            displayed whenever the :meth:`display_state` method is called, defaults to every
            player.
        enable_animation (bool): If set to True, will play animations, default True.
        no_wait (bool): If set to True, disables waiting mechanisms and will not block.
    Attributes:
        game (TexasHoldEm, optional): The game object to attach to, all methods will
            default to this game. (Not necessary if only showing the history)
        visible_players (Iterable[int], optional): The players whose cards should be
            displayed whenever the :meth:`display_state` method is called, defaults to every
            player.
        enable_animation (bool): If set to True, will play animations, default True.
        no_wait (bool): If set to True, disables waiting mechanisms and will not block.

    """

    _action_patterns = (
        (r"^all(\-|\s|_)?in$", ActionType.ALL_IN),
        (r"^call$", ActionType.CALL),
        (r"^check$", ActionType.CHECK),
        (r"^fold$", ActionType.FOLD),
        (r"^raise (to )?([0-9]+)$", ActionType.RAISE)
    )

    def __init__(self,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._command_patterns = (
            (r'^exit|quit$', self._exit_handler),
        )

        # init curses
        self.main_block = _Block(name="Main Window",
                                 window=curses.initscr())

        # handle resize gracefully
        if not _IS_WINDOWS:
            signal.signal(signal.SIGWINCH,
                          lambda signals, frame: (self.refresh(), self.main_block.window.getch()))

        # cleanup before exit
        signal.signal(signal.SIGINT, lambda signals, frame: self._exit_handler())

        # hide screen until called
        if self.game:
            self.refresh()
        self.hide()

    def _exit_handler(self):
        """
        Exit handler snippet
        """
        self.hide()
        sys.exit(2)

    @deprecated(version="0.7.0",
                reason="Use the :meth:`set_visible_players` method instead. "
                       "This function will be removed in version 1.0.0.")
    def set_player_ids(self, ids: Iterable[int]):
        """
        Make the given players' cards visible.

        Arguments:
            ids (Iterable[int]): The players whose cards should be visible when the
                :meth:`display_state` method is called.

        """
        self.visible_players = list(ids)

    @deprecated(version="0.7.0",
                reason="Use the :meth:`display_action` method instead. "
                       "This function will be removed in version 1.0.0.")
    def print_action(self, id: int, action: ActionType, val: Optional[int] = None):
        # pylint: disable=redefined-builtin,unused-argument
        """
        Display the most recent action

        """
        self._display_action(player_id=id, action=action)

    @deprecated(version="0.7.0",
                reason="Use the :meth:`display_state` method instead. "
                       "This function will be removed in version 1.0.0.")
    def print_state(self, poker_game: TexasHoldEm):
        """
        Display the state of the game.

        """
        self.game = poker_game
        self.refresh()
        return self.display_state()

    def _capture_string(self) -> str:
        """
        Helper function for the :meth:`accept_input` method. Captures an inputed
        string and handles backspaces, newlines, resize key strokes, etc.

        Returns:
            str: The captured string ended by a newline
        """
        rows, _ = self.main_block.window.getmaxyx()
        string = ""
        i = len(_PROMPT)
        while True:
            ord_ = self.main_block.window.getch(rows - 1, i)

            if ord_ == _BACKSPACE:
                # Delete the backspace char
                self.main_block.window.delch(rows - 1, i + 1)
                self.main_block.window.delch(rows - 1, i)

                # Don't delete the prompt
                if i <= len(_PROMPT):
                    continue

                # delete the previous char
                self.main_block.window.delch(rows - 1, i - 1)
                i -= 1
                string = string[:-1]

            # stop string collection on newline
            elif ord_ == _NEWLINE:
                break

            # Windows Compatibility, need a workaround for SIGINT
            # For now, allow users to ctrl+c in the input phase
            elif _IS_WINDOWS and ord_ == _CTRL_C:
                self._exit_handler()

            # add to the string
            else:
                i += 1

                try:
                    string += chr(ord_)
                except ValueError as err:
                    # fail silently (don't want to echo) but preserve
                    # stack trace
                    raise Ignore() from err

        return string.strip()

    @preflight(prerun=lambda self: self.refresh())
    def accept_input(self) -> Tuple[ActionType, Optional[int]]:
        curses.echo(True)
        curses.curs_set(1)

        string = self._capture_string()

        curses.echo(False)
        curses.curs_set(0)

        self.main_block.get_block('INPUT').erase()
        self.main_block.refresh()

        string = string.lower().strip()

        # empty string noop
        if not string:
            raise Ignore()

        # special functions
        for pattern, func in self._command_patterns:
            if re.match(pattern, string):
                func()
                raise Ignore()

        # actions
        for pattern, action_type in self._action_patterns:
            match = re.match(pattern, string)

            if match:
                total = None
                if action_type == ActionType.RAISE:
                    total = int(match.group(2))

                # erase any errors
                self.main_block.get_block('ERROR').erase()

                return action_type, total

        raise ValueError(f"Could not parse '{string}'")

    def _recalculate_object_blocks(self):
        """
        Recalculates every block location and places them there
        (does not fill with with content, only places them)

        """
        rows, cols = self.main_block.window.getmaxyx()

        # Place input box on the bottom left
        self.main_block.new_block(
            'INPUT',
            *_PROMPT_BLOCK_SIZE,
            (rows - _PROMPT_BLOCK_SIZE[0]),
            0
        )
        input_size = self.main_block.get_block('INPUT').window.getmaxyx()

        # Place error box on the bottom left right above the input
        self.main_block.new_block(
            'ERROR',
            *_ERROR_BLOCK_SIZE,
            (rows - input_size[0] - _ERROR_BLOCK_SIZE[0]),
            0
        )

        player_ellipse = _Ellipse(
            major=(cols / 2) * _PLAYER_ELLIPSE_SIZE_FACTOR,
            minor=(rows / 2) * _PLAYER_ELLIPSE_SIZE_FACTOR,
            center=(cols / 2 + _TABLE_ELLIPSE_OFFSET[0], rows / 2 + _TABLE_ELLIPSE_OFFSET[1])
        )
        player_bet_ellipse = _Ellipse(
            major=(cols / 2) * _PLAYER_BET_ELLIPSE_SIZE_FACTOR,
            minor=(rows / 2) * _PLAYER_BET_ELLIPSE_SIZE_FACTOR,
            center=(cols / 2 + _TABLE_ELLIPSE_OFFSET[0], rows / 2 + _TABLE_ELLIPSE_OFFSET[1])
        )

        # Place player windows in an ellipse with player 0 at the bottom of the screen
        # and continuing clockwise.
        start_rad = math.pi / 2
        rad_per_player = ((2 * math.pi) / self.game.max_players)
        for player_id in range(self.game.max_players):
            rad = start_rad + rad_per_player * player_id
            y, x = player_ellipse.point_yx(rad)
            self.main_block.new_block(
                f'PLAYER_INFO_{player_id}',
                *_PLAYER_BLOCK_SIZE,
                round(y) - _PLAYER_BLOCK_SIZE[0] // 2,
                round(x) - _PLAYER_BLOCK_SIZE[1] // 2)

            y, x = player_bet_ellipse.point_yx(rad)
            self.main_block.new_block(
                f'PLAYER_CHIPS_{player_id}',
                *_PLAYER_BET_BLOCK_SIZE,
                round(y) - _PLAYER_BET_BLOCK_SIZE[0] // 2,
                round(x) - _PLAYER_BET_BLOCK_SIZE[1] // 2)

        # Place the board and pots in the center
        self.main_block.new_block(
            'BOARD',
            *_BOARD_BLOCK_SIZE,
            (rows - _BOARD_BLOCK_SIZE[0]) // 2 + _TABLE_ELLIPSE_OFFSET[1],
            (cols - _BOARD_BLOCK_SIZE[1]) // 2 + _TABLE_ELLIPSE_OFFSET[0])

        # Place history on the right
        self.main_block.new_block(
            'HISTORY',
            round(rows * _HISTORY_BLOCK_SIZE_FACTOR),
            _HISTORY_BLOCK_SIZE[1],
            rows - round(rows * _HISTORY_BLOCK_SIZE_FACTOR),
            cols - _HISTORY_BLOCK_SIZE[1]
        )

        # version block above history
        self.main_block.new_block(
            'VERSION',
            *_VERSION_BLOCK_SIZE,
            0,
            cols - _VERSION_BLOCK_SIZE[1]
        )

    def _player_block(self, player_id: int) -> list[str]:
        """
        Arguments:
            player_id (int): The player id
        Returns:
            list[str]: The content for the given player

        """
        block = []

        block.extend([f"Player {player_id}",
                      f"{self.game.players[player_id].state.name}",
                      f"Chips: {self.game.players[player_id].chips}"])

        for blind, blind_str in ((self.game.btn_loc, "Button"),
                                 (self.game.sb_loc, "Small Blind"),
                                 (self.game.bb_loc, "Big Blind")):
            if player_id == blind:
                block.append(blind_str)
                break

        if self.game.players[player_id].state != PlayerState.SKIP:
            in_pot = list(self.game.in_pot_iter())
            if (player_id in self.visible_players
                    or (self.game.hand_phase == HandPhase.SETTLE
                        and len(in_pot) > 1
                        and player_id in in_pot)):
                block.append(
                    card.card_list_to_pretty_str(self.game.get_hand(player_id))
                )
            else:
                block.append('[ * ] [ * ]')

        return block

    def _player_bet_block(self, player_id: int) -> list[str]:
        """
        Arguments:
            player_id (int): The player id
        Returns:
            list[str]: The player bet amount content

        """
        return [f"Bet: {self.game.player_bet_amount(player_id)}"]

    @staticmethod
    def _version_block() -> list[str]:
        """
        Returns:
            list[str]: The version block content

        """
        return [f"texaholdem: v{version('texasholdem')}"]

    def _history_block(self) -> list[str]:
        """
        The history block includes headers for each hand phase, action callouts
        for each player during their turn, and how many chips for winners.

        Returns:
            list[str]: The content for the history

        """
        history_rows, history_cols = self.main_block.get_block('HISTORY').window.getmaxyx()
        history_rows, history_cols = history_rows - 2, history_cols - 3   # for the border / newline
        history_border = '-' * history_cols
        block = deque(maxlen=history_rows)

        block.append(f'Hand #{self.game.num_hands}')
        for hand_phase in (HandPhase.PREFLOP,
                           HandPhase.FLOP,
                           HandPhase.TURN,
                           HandPhase.RIVER):
            if hand_phase in self.game.hand_history:
                block.append(history_border)
                block.append(hand_phase.name)
                block.append(history_border)
                for action in self.game.hand_history[hand_phase].actions:
                    block.append(str(action))

        if HandPhase.SETTLE in self.game.hand_history:
            block.append(history_border)
            block.append(HandPhase.SETTLE.name)
            block.append(history_border)
            block.extend(str(self.game.hand_history[HandPhase.SETTLE]).split('\n'))

        return list(block)

    def _board_block(self) -> list[str]:
        """
        The board block includes the board cards and the pots.

        Returns:
            list[str]: The content for the board and for the pots

        """
        return [f"Board: {card.card_list_to_pretty_str(self.game.board)}",
                "",
                *(f"Pot {i}: {pot.get_amount()} ({pot.get_total_amount() - pot.get_amount()})"
                  for i, pot in enumerate(self.game.pots))]

    def _paint_table_ring(self):
        """
        Paints the table ellipse directly the main window.

        """
        # paint table
        rows, cols = self.main_block.window.getmaxyx()
        table_ellipse = _Ellipse(
            major=(cols / 2) * _TABLE_ELLIPSE_SIZE_FACTOR,
            minor=(rows / 2) * _TABLE_ELLIPSE_SIZE_FACTOR,
            center=(cols / 2 + _TABLE_ELLIPSE_OFFSET[0], rows / 2 + _TABLE_ELLIPSE_OFFSET[1])
        )

        for step in range(1, _TABLE_STEPS_RESOLUTION):
            rad = ((2 * math.pi) / _TABLE_STEPS_RESOLUTION) * step
            y, x = table_ellipse.point_yx(rad)
            self.main_block.window.addstr(
                *self.main_block.bound_coords(round(y), round(x)),
                table_ellipse.char_at(rad)
            )

    def refresh(self):
        """
        Refreshes the display

        """
        self.main_block.stash_state()
        self.main_block.window.clear()

        x, y = shutil.get_terminal_size()
        curses.resizeterm(y, x)

        self._paint_table_ring()
        self._recalculate_object_blocks()

        self.main_block.pop_state()
        self.main_block.refresh()

    def hide(self):
        curses.endwin()

    def display_state(self):
        # paint board
        self.main_block.blocks['BOARD'].add_content(
            content=self._board_block()
        )

        # paint players
        for player_id in range(self.game.max_players):
            self.main_block.blocks[f'PLAYER_INFO_{player_id}'].add_content(
                content=self._player_block(player_id),
                border=player_id == self.game.current_player
            )

            self.main_block.blocks[f'PLAYER_CHIPS_{player_id}'].add_content(
                content=self._player_bet_block(player_id)
            )

        # history
        self.main_block.blocks['HISTORY'].add_content(
            content=self._history_block(),
            align=_Align.BOTTOM,
            border=True,
            wrap_line=True
        )

        # version
        self.main_block.blocks['VERSION'].add_content(
            content=self._version_block()
        )
        self.main_block.refresh()

    def prompt_input(self, preamble: Optional[list[str]] = None):
        if preamble is None:
            preamble = [f"Player {self.game.current_player}\'s turn"]

        self.main_block.get_block('INPUT').erase()
        self.main_block.get_block('INPUT').add_content(
            [*preamble,
             _PROMPT],
            align=_Align.BOTTOM,
            justify=_Justify.LEFT
        )
        self.main_block.refresh()

    def display_error(self, error: str):
        self.main_block.get_block('ERROR').erase()
        self.main_block.get_block('ERROR').add_content(
            [error],
            align=_Align.BOTTOM,
            justify=_Justify.LEFT
        )
        self.main_block.refresh()

    @handle(handler=lambda exc: logger.info("Skipping because animation is disabled"),
            exc_type=Ignore)
    @preflight(prerun=lambda self, *args, **kwargs: raise_if(Ignore(), not self.enable_animation))
    def _display_action(self, player_id: int, action: ActionType):
        """
        Animates the chip movement for raise and call actions.

        """
        curses.curs_set(0)

        if action in (ActionType.RAISE, ActionType.CALL):
            player_y, player_x = \
                self.main_block.get_block(f'PLAYER_INFO_{player_id}').window.getbegyx()
            bet_y, bet_x = \
                self.main_block.get_block(f'PLAYER_CHIPS_{player_id}').window.getbegyx()

            start_y, start_x = (player_y + _PLAYER_BLOCK_SIZE[0] // 2,
                                player_x + _PLAYER_BLOCK_SIZE[1] // 2)
            end_y, end_x = (bet_y + _PLAYER_BET_BLOCK_SIZE[0] // 2,
                            bet_x + _PLAYER_BET_BLOCK_SIZE[1] // 2)

            y, x = start_y, start_x
            tick_y, tick_x = ((end_y - start_y) / _ACTION_STEPS,
                              (end_x - start_x) / _ACTION_STEPS)

            self.main_block.new_block('ACTION',
                                      *_ACTION_BLOCK_SIZE,
                                      round(y),
                                      round(x)).add_content(['*'])

            while (round(y), round(x)) != (end_y, end_x):
                y += tick_y
                x += tick_x

                self.main_block.get_block('ACTION').erase()
                self.main_block.new_block('ACTION',
                                          *_ACTION_BLOCK_SIZE,
                                          round(y),
                                          round(x)).add_content(['*'])
                curses.napms(_ACTION_SLEEP_MS)
                self.refresh()

            self.main_block.blocks.pop('ACTION').window.erase()

    def display_action(self):
        if not self.game.hand_history.combined_actions():
            return
        player_action = self.game.hand_history.combined_actions()[-1]
        player_id, action = player_action.player_id, player_action.action_type
        self._display_action(player_id, action)

    def display_win(self):
        old_visible_players = self.visible_players

        # in the settle phase, players going to showdown show cards
        extras = set(self.game.in_pot_iter())
        extras = extras if len(extras) > 1 else []  # don't out players win without contest
        self.set_visible_players(set(self.visible_players).union(extras))

        self.display_state()
        self.main_block.refresh()

        self.wait_until_prompted()

        self.visible_players = old_visible_players

    @handle(handler=lambda exc: logger.info("Skipping because no_wait is True"),
            exc_type=Ignore)
    @preflight(prerun=lambda self: raise_if(Ignore(), self.no_wait))
    def wait_until_prompted(self):
        self.prompt_input(preamble=["Press enter to continue"])
        curses.curs_set(0)
        self.main_block.refresh()
        self.main_block.window.getstr()



from typing import Callable, Any, TypeVar, Type, Tuple

from functools import wraps

_F = TypeVar("_F", bound=Callable)
_T = TypeVar("_T")
_R = TypeVar("_R")
_E = TypeVar("_E", bound=Type[Exception])


def check_raise(exc_type: Type[Exception]):
    """
    Decorator that turns a function that returns a bool and message into
    a function that returns a bool and optionally throws an error.

    Injects argument throws (bool): default False.

    Example:
        You can use the decorator like so::

            @check_raise(ValueError)
            def validate(arg1):
                return False, "test"

            validate("arg1")                # will return False
            validate("arg1", throws=True)   # will raise ValueError("test")

    Arguments:
        exc_type (Type[Exception]): The Exception type to throw.
    """

    def decorator(func: Callable[[_T], Tuple[bool, str]]) -> Callable[[_T], bool]:

        @wraps(func)
        def inner(*args, throws=False, **kwargs):
            ret, msg = func(*args, **kwargs)
            if not ret and throws:
                raise exc_type(msg)
            return ret

        return inner

    return decorator


def handle(handler: Callable[[_E], Any],
           exc_type: _E = Exception):
    """
    Decorator that wraps the entire function in a try-except statement
    that catches the given exc_type and handles it with the given handler.

    Arguments:
        exc_type (Type[Exception]): The exception type to handle
        handler: Function that handles the exc_type exception

    """

    def decorator(func: _F) -> _F:

        @wraps(func)
        def inner(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except exc_type as exc:             # # pylint: disable=broad-except
                return handler(exc)

        return inner

    return decorator


def preflight(prerun: Callable[[_T], Any]):
    """
    Decorator that runs the given function with the same parameters of the
    decorated function beforehand.

    Arguments:
        prerun: Function that runs before the decorated function, takes the same arguments.
    """

    def decorator(func: Callable[[_T], _R]) -> Callable[[_T], _R]:

        @wraps(func)
        def inner(*args, **kwargs):
            prerun(*args, **kwargs)
            return func(*args, **kwargs)

        return inner

    return decorator


def raise_if(exc: Exception, condition: bool):
    """
    Throws the given exception if condition is True. Useful to throw errors
    in lambda statements.

    Arguments:
        exc (Exception): The exception instance to throw
        condition (bool): If true, will raise the given error

    """
    if condition:
        raise exc



"""
The lookup table module keeps the books on all possible hand strengths.
We construct the table once on import and reference it repeatedly.

Number of Distinct Hand Values::

    Straight Flush   10
    Four of a Kind   156      [(13 choose 2) * (2 choose 1)]
    Full Houses      156      [(13 choose 2) * (2 choose 1)]
    Flush            1277     [(13 choose 5) - 10 straight flushes]
    Straight         10
    Three of a Kind  858      [(13 choose 3) * (3 choose 1)]
    Two Pair         858      [(13 choose 3) * (3 choose 2)]
    One Pair         2860     [(13 choose 4) * (4 choose 1)]
    High card      + 1277     [(13 choose 5) - 10 straights]
    -------------------------
    TOTAL            7462

Here we create a lookup table which maps:

    - 5 card hand's unique prime product -> rank in range [1, 7462]

Example:
    - Royal flush (best hand possible) -> 1
    - 7-5-4-3-2 unsuited (worst hand possible) -> 7462

"""

from typing import Dict
import itertools

from texasholdem.card import card
from texasholdem.card.card import Card


class LookupTable:
    # pylint: disable=too-few-public-methods
    """
    Attributes:
        flush_lookup (Dict[int, int]): map from prime-product to rank for suited cards
        unsuited_lookup (Dict[int, int]): map from prime-product to rank for unsuited cards

    """

    MAX_STRAIGHT_FLUSH = 10
    MAX_FOUR_OF_A_KIND = 166
    MAX_FULL_HOUSE = 322
    MAX_FLUSH = 1599
    MAX_STRAIGHT = 1609
    MAX_THREE_OF_A_KIND = 2467
    MAX_TWO_PAIR = 3325
    MAX_PAIR = 6185
    MAX_HIGH_CARD = 7462

    MAX_TO_RANK_CLASS = {
        MAX_STRAIGHT_FLUSH: 1,
        MAX_FOUR_OF_A_KIND: 2,
        MAX_FULL_HOUSE: 3,
        MAX_FLUSH: 4,
        MAX_STRAIGHT: 5,
        MAX_THREE_OF_A_KIND: 6,
        MAX_TWO_PAIR: 7,
        MAX_PAIR: 8,
        MAX_HIGH_CARD: 9
    }

    RANK_CLASS_TO_STRING = {
        1: "Straight Flush",
        2: "Four of a Kind",
        3: "Full House",
        4: "Flush",
        5: "Straight",
        6: "Three of a Kind",
        7: "Two Pair",
        8: "Pair",
        9: "High card"
    }

    def __init__(self):
        # create dictionaries
        self.flush_lookup: Dict[int, int] = {}
        self.unsuited_lookup: Dict[int, int] = {}

        # create the lookup table in piecewise fashion
        self._flushes()  # this will call straights and high card method,
        # we reuse some of the bit sequences
        self._multiples()

    def _flushes(self):
        """
        Straight flushes and flushes.

        Lookup is done on 13 bit integer (2^13 > 7462):
        xxxbbbbb bbbbbbbb => integer hand index

        """

        # straight flushes in rank order
        straight_flushes = [
            7936,  # int('0b1111100000000', 2), # royal flush
            3968,  # int('0b111110000000', 2),
            1984,  # int('0b11111000000', 2),
            992,  # int('0b1111100000', 2),
            496,  # int('0b111110000', 2),
            248,  # int('0b11111000', 2),
            124,  # int('0b1111100', 2),
            62,  # int('0b111110', 2),
            31,  # int('0b11111', 2),
            4111  # int('0b1000000001111', 2) # 5 high
        ]

        # now we'll dynamically generate all the other
        # flushes (including straight flushes)
        flushes = []
        gen = LookupTable._get_lexographically_next_bit_sequence(int('0b11111', 2))

        # 1277 = number of high card
        # 1277 + len(str_flushes) is number of hands with all card unique rank
        for _ in range(1277 + len(straight_flushes) - 1):  # we also iterate over SFs
            # pull the next flush pattern from our generator
            flush = next(gen)

            # if this flush matches perfectly any
            # straight flush, do not add it
            not_sf = True
            for straight_flush in straight_flushes:
                # if f XOR sf == 0, then bit pattern
                # is same, and we should not add
                if not flush ^ straight_flush:
                    not_sf = False

            if not_sf:
                flushes.append(flush)

        # we started from the lowest straight pattern, now we want to start ranking from
        # the most powerful hands, so we reverse
        flushes.reverse()

        # now add to the lookup map:
        # start with straight flushes and the rank of 1
        # since theyit is the best hand in poker
        # rank 1 = Royal Flush!
        rank = 1
        for straight_flush in straight_flushes:
            prime_product = card.prime_product_from_rankbits(straight_flush)
            self.flush_lookup[prime_product] = rank
            rank += 1

        # we start the counting for flushes on max full house, which
        # is the worst rank that a full house can have (2,2,2,3,3)
        rank = LookupTable.MAX_FULL_HOUSE + 1
        for flush in flushes:
            prime_product = card.prime_product_from_rankbits(flush)
            self.flush_lookup[prime_product] = rank
            rank += 1

        # we can reuse these bit sequences for straights
        # and high card since they are inherently related
        # and differ only by context
        self._straight_and_highcards(straight_flushes, flushes)

    def _straight_and_highcards(self, straights, highcards):
        """
        Unique five card sets. Straights and highcards.

        Reuses bit sequences from flush calculations.
        """
        rank = LookupTable.MAX_FLUSH + 1

        for straight in straights:
            prime_product = card.prime_product_from_rankbits(straight)
            self.unsuited_lookup[prime_product] = rank
            rank += 1

        rank = LookupTable.MAX_PAIR + 1
        for high_card in highcards:
            prime_product = card.prime_product_from_rankbits(high_card)
            self.unsuited_lookup[prime_product] = rank
            rank += 1

    def _multiples(self):
        # pylint: disable=too-many-locals
        """
        Pair, Two Pair, Three of a Kind, Full House, and 4 of a Kind.
        """
        backwards_ranks = range(len(Card.INT_RANKS) - 1, -1, -1)

        # 1) Four of a Kind
        rank = LookupTable.MAX_STRAIGHT_FLUSH + 1

        # for each choice of a set of four rank
        for i in backwards_ranks:

            # and for each possible kicker rank
            kickers = list(backwards_ranks[:])
            kickers.remove(i)
            for k in kickers:
                product = Card.PRIMES[i] ** 4 * Card.PRIMES[k]
                self.unsuited_lookup[product] = rank
                rank += 1

        # 2) Full House
        rank = LookupTable.MAX_FOUR_OF_A_KIND + 1

        # for each three of a kind
        for i in backwards_ranks:

            # and for each choice of pair rank
            pair_ranks = list(backwards_ranks[:])
            pair_ranks.remove(i)
            for pair_rank in pair_ranks:
                product = Card.PRIMES[i] ** 3 * Card.PRIMES[pair_rank] ** 2
                self.unsuited_lookup[product] = rank
                rank += 1

        # 3) Three of a Kind
        rank = LookupTable.MAX_STRAIGHT + 1

        # pick three of one rank
        for b_rank in backwards_ranks:

            kickers = list(backwards_ranks[:])
            kickers.remove(b_rank)

            for kickers in itertools.combinations(kickers, 2):
                card1, card2 = kickers
                product = Card.PRIMES[b_rank] ** 3 * Card.PRIMES[card1] * Card.PRIMES[card2]
                self.unsuited_lookup[product] = rank
                rank += 1

        # 4) Two Pair
        rank = LookupTable.MAX_THREE_OF_A_KIND + 1

        for two_pair in itertools.combinations(backwards_ranks, 2):

            pair1, pair2 = two_pair
            kickers = list(backwards_ranks[:])
            kickers.remove(pair1)
            kickers.remove(pair2)
            for kicker in kickers:
                product = Card.PRIMES[pair1] ** 2 * Card.PRIMES[pair2] ** 2 * Card.PRIMES[kicker]
                self.unsuited_lookup[product] = rank
                rank += 1

        # 5) Pair
        rank = LookupTable.MAX_TWO_PAIR + 1

        # choose a pair
        for pairrank in backwards_ranks:

            kickers = list(backwards_ranks[:])
            kickers.remove(pairrank)

            for kickers in itertools.combinations(kickers, 3):
                kicker1, kicker2, kicker3 = kickers
                product = Card.PRIMES[pairrank] ** 2 * Card.PRIMES[kicker1] \
                    * Card.PRIMES[kicker2] * Card.PRIMES[kicker3]
                self.unsuited_lookup[product] = rank
                rank += 1

    @staticmethod
    def _get_lexographically_next_bit_sequence(bits):
        """
        Bit hack from here:
        http://www-graphics.stanford.edu/~seander/bithacks.html#NextBitPermutation

        Generator even does this in poker order rank
        so no need to sort when done!
        """
        xbits = (bits | (bits - 1)) + 1
        lexo_next = xbits | ((((xbits & -xbits) // (bits & -bits)) >> 1) - 1)
        yield lexo_next
        while True:
            xbits = (lexo_next | (lexo_next - 1)) + 1
            lexo_next = xbits | ((((xbits & -xbits) // (lexo_next & -lexo_next)) >> 1) - 1)
            yield lexo_next


LOOKUP_TABLE = LookupTable()
"""
The lookup table that is created when imported
"""



from __future__ import annotations

from enum import Enum
from dataclasses import dataclass


@dataclass(frozen=True)
class _HandPhase:
    """
    Hand phase
    """
    new_cards: int
    next_phase: str


class HandPhase(Enum):
    """An enum representing the phases of each hand which includes the well-known

        - :obj:`HandPhase.PREFLOP`
        - :obj:`HandPhase.FLOP`
        - :obj:`HandPhase.TURN`
        - :obj:`HandPhase.RIVER`

    In addition to two new phases

        - :obj:`HandPhase.PREHAND`
        - :obj:`HandPhase.SETTLE`

    which are used for book-keeping.

    """

    PREHAND = _HandPhase(0, "PREFLOP")
    """
    In this phase, players sit out if requested, players
    rejoin if requested, blinds are moved and posted, and card
    are dealt.
    """

    PREFLOP = _HandPhase(0, "FLOP")
    """
    The first betting round of the game. Players have two cards with no
    communal cards yet.
    """

    FLOP = _HandPhase(3, "TURN")
    """
    The second betting round of the game. Three communal cards come outs.
    """

    TURN = _HandPhase(1, "RIVER")
    """
    The third betting round of the game. One more communal card comes out.
    """

    RIVER = _HandPhase(1, "SETTLE")
    """
    The fourth and final betting round of the game. One more communal cards come out.
    """

    SETTLE = _HandPhase(0, "PREHAND")
    """
    If the hand ended early in a previous round, the rest of the communal cards come out
    to total 5. Winners are decided per pot based on hand strength and rewarded chips.
    """

    def next_phase(self) -> HandPhase:
        """
        Returns:
            HandPhase: The next HandPhase after this one

        """
        return HandPhase[self.value.next_phase]

    def new_cards(self) -> int:
        """
        Returns:
            int: The number of new cards to add to the board

        """
        return self.value.new_cards



"""The history module includes various dataclasses intended
to keep the history of the game to be able to replay hands and
export/import notation of the game.

Texas Hold Em Notation Conventions:

    - The button is assigned ID 0
    - Each action is a tuple :code:`(player_id, action_type, total)`
    - The winner section is a per-pot tuple :code:`(Pot id, amount, best rank, winner ids)`

"""
from __future__ import annotations
from typing import Optional, Union, Tuple, List, Dict
from dataclasses import dataclass
from pathlib import Path
import os

from texasholdem.game.action_type import ActionType
from texasholdem.card.card import Card
from texasholdem.game.hand_phase import HandPhase
from texasholdem.evaluator import evaluator


FILE_EXTENSION = 'pgn'
"""
The extension for the game notation file.

"""


class HistoryImportError(Exception):
    """
    The history classes will throw this error if it cannot import the given PGN file.

    """


@dataclass()
class PrehandHistory:
    """
    History of the Prehand Phase

    """
    btn_loc: int
    """
    The id of the player with the button
    """
    big_blind: int
    """
    The id of the player with the big blind
    """
    small_blind: int
    """
    The id of the player with the small blind
    """
    player_chips: Dict[int, int]
    """
    The number of chips for each player
    """
    player_cards: Dict[int, List[Card]]
    """
    The cards for each player
    """

    def to_string(self, canon_ids: Dict[int, int]) -> str:
        """
        Arguments:
            canon_ids (Dict[int, int]): Map of old_id -> new_id where the new btn_loc is 0
        Returns:
            str: The string representation of the prehand history: blind sizes, chips, and cards

        """
        str_player_cards = {i: [str(card) for card in self.player_cards[i]] for i in canon_ids}

        return f"Big Blind: {self.big_blind}\n" \
               f"Small Blind: {self.small_blind}\n" \
               f"Player Chips: {','.join(str(self.player_chips[i]) for i in canon_ids)}\n" \
               f"Player Cards: " \
               f"{','.join(['[' + ' '.join(str_player_cards[i]) + ']' for i in canon_ids])}"

    @staticmethod
    def from_string(string: str) -> PrehandHistory:
        """
        Reverse of to_string()

        Arguments:
            string (str): The string as returned from to_string()
        Returns:
            PrehandHistory: The prehand history as represented by the string
        Raises:
            HistoryImportError: If missing information or a mismatch between cards and chips

        """
        big_blind, small_blind, chips_str, cards_str = string.split('\n')
        _, big_blind = big_blind.split(': ')
        _, small_blind = small_blind.split(': ')

        _, chips_str = chips_str.split(': ')
        player_chips = [int(chip_str) for chip_str in chips_str.split(',')]
        num_players = len(player_chips)

        _, cards_str = cards_str.split(': ')
        cards_data = cards_str.split(',')
        cards_data = [card_data.strip('[]').split(' ') for card_data in cards_data]
        player_cards = [[Card(c1), Card(c2)] for c1, c2 in cards_data]

        if len(player_chips) != len(player_cards):
            raise HistoryImportError(f"Mismatch number of player chips ({len(player_chips)}) "
                                     f"and player cards ({len(player_cards)})")

        return PrehandHistory(0,
                              int(big_blind),
                              int(small_blind),
                              dict(zip(range(num_players), player_chips)),
                              dict(zip(range(num_players), player_cards)))


@dataclass()
class PlayerAction:
    """
    History of a Player Action

    """
    player_id: int
    """
    The player id
    """
    action_type: ActionType
    """
    The action type
    """
    total: Optional[int] = None
    """
    The total raise amount
    """
    value: Optional[int] = None
    """
    The amount raised
    """

    def to_string(self, canon_ids: Dict[int, int]) -> str:
        """
        Arguments:
            canon_ids (Dict[int, int]): Map of old_id -> new_id where the new btn_loc is 0
        Returns:
            str: The string representation of a player action: id, action, amount
        """
        string = f"({canon_ids[self.player_id]},{self.action_type.name}"
        if self.total is not None and self.total > 0:
            string += "," + str(self.total)
        string += ")"

        return string

    @staticmethod
    def from_string(string: str) -> PlayerAction:
        """
        Reverse of to_string()

        Arguments:
            string (str): The string as returned from to_string()
        Returns:
            PlayerAction: The player action as represented by the string
        """
        string = string.strip().strip('()')
        data = string.split(",")
        player_id, action_type = int(data[0]), ActionType[data[1]]
        total = None if len(data) <= 2 else int(data[2])
        return PlayerAction(player_id=player_id, action_type=action_type, total=total)

    def __str__(self) -> str:
        return f"Player {self.player_id} {self.action_type.name}" \
               f"{f' to {self.total}' if self.action_type == ActionType.RAISE else ''}"


@dataclass()
class BettingRoundHistory:
    """
    History of a Betting Round

    """
    new_cards: List[Card]
    """
    The new cards that were added this round
    """
    actions: List[PlayerAction]
    """
    A list of PlayerActions
    """

    def to_string(self, canon_ids: Dict[int, int]) -> str:
        """
        Arguments:
            canon_ids (Dict[int, int]): Map of old_id -> new_id where the new btn_loc is 0
        Returns:
            str: The string representation of the betting round history:
                new cards revealed, ordered list of actions
        """
        new_cards = f"New Cards: [{','.join(str(card) for card in self.new_cards)}]"
        counts = {}
        orbits = {}

        for action in self.actions:
            counts[action.player_id] = counts.get(action.player_id, 0) + 1
            min_count = max(counts.values())

            if min_count not in orbits:
                orbits[min_count] = []
            orbits[min_count].append(action.to_string(canon_ids))

        orbit_lines = [f"{orbit_num}. " +
                       ";".join(orbit_line) for orbit_num, orbit_line in orbits.items()]
        action_str = "\n".join(orbit_lines)
        return new_cards + "\n" + action_str

    @staticmethod
    def from_string(string: str) -> BettingRoundHistory:
        """
        Reverse of to_string()

        Arguments:
            string (str): The string as returned from to_string()
        Returns:
            BettingRoundHistory: The betting round as represented by the string
        """
        data = string.split('\n')
        card_str, data = data[0], data[1:]
        _, card_str = card_str.split('[')
        card_str, _ = card_str.split(']')

        if card_str:
            new_cards = [Card(string) for string in card_str.split(',')]
        else:
            new_cards = []

        action_str = ""
        for i, action_line in enumerate(data, 1):
            _, action_line = action_line.split(". ")
            action_str += action_line
            if i != len(data):
                action_str += ';'

        actions = [PlayerAction.from_string(string) for string in action_str.split(";")]

        return BettingRoundHistory(new_cards, actions)


@dataclass()
class SettleHistory:
    """
    History of the Settle Phase

    """
    new_cards: List[Card]
    """
    The new cards that were added this round
    """
    pot_winners: Dict[int, Tuple[int, int, List[int]]]
    """
    A map from pot_id to a tuple of winner data :code:`amount, best rank, list of winners`
    """

    def to_string(self, canon_ids: Dict[int, int]) -> str:
        """
        Arguments:
            canon_ids (Dict[int, int]): Map of old_id -> new_id where the new btn_loc is 0
        Returns:
            str: The string representation of the settle history: new cards revealed,
                and the winners per pot: (pot_id, total_amount, best_rank, winners list)
        """
        pot_lists = []
        for pot_id, (amount, best_rank, winners_list) in self.pot_winners.items():
            pot_lists.append((pot_id, amount, best_rank, [canon_ids[winner]
                                                          for winner in winners_list]))
        pot_strs = [f'(Pot {pot_id},{amount},{best_rank},{str(winners_list)})'
                    for pot_id, amount, best_rank, winners_list in pot_lists]
        return f"New Cards: [{','.join(str(card) for card in self.new_cards)}]\n" \
               f"Winners: {';'.join(pot_strs)}"

    @staticmethod
    def from_string(string: str) -> SettleHistory:
        """
        Reverse of to_string()

        Arguments:
            string (str): The string as returned from to_string()
        Returns:
            SettleHistory: The settle history as represented by the string
        """
        cards_str, winners_str = string.split('\n')
        _, cards_str = cards_str.split('[')
        cards_str, _ = cards_str.split(']')

        if cards_str:
            new_cards = [Card(string) for string in cards_str.split(',')]
        else:
            new_cards = []

        _, winners_str = winners_str.split(': ')
        pot_winners_data = [winner_str.strip('()')
                            .replace(' ', '')
                            .replace('[', '')
                            .replace(']', '')
                            .split(',')
                            for winner_str in winners_str.split(';')]
        pot_winners_data = [(data[0], data[1], data[2], data[3:])
                            for data in pot_winners_data]
        pot_winners = {int(pot_name[(pot_name.find('Pot') + 3):]):
                       (int(amount), int(best_rank), [int(winner) for winner in winners_list])
                       for pot_name, amount, best_rank, winners_list in pot_winners_data}
        return SettleHistory(new_cards, pot_winners)

    def __str__(self) -> str:
        pot_lists = []
        for pot_id, (amount, best_rank, winners_list) in self.pot_winners.items():
            winners_list = [str(winner) for winner in winners_list]
            single = len(winners_list) > 0
            rank_str = f" with a {evaluator.rank_to_string(best_rank)}" if best_rank != -1 else ""

            pot_lists.append(f"Player{'s' if not single else ''} "
                             f"{', '.join(winners_list)} {'wins' if single else 'splits'} "
                             f"Pot {pot_id} (Chips {amount}){rank_str}")
        return "\n".join(pot_lists)


@dataclass()
class History:
    """
    History class of a hand of TexasHoldEm. Includes one history item for each HandPhase.
    In total, this constitutes a minimal amount of information necessary to replay a hand.

    This class also includes :meth:`to_string()` and :meth:`from_string()` methods which
    provides ways to write / read human-readable information to / from files.

    """
    prehand: PrehandHistory = None
    """
    The Prehand History
    """
    preflop: BettingRoundHistory = None
    """
    The Preflop History
    """
    settle: SettleHistory = None
    """
    The Settle History
    """
    flop: Optional[BettingRoundHistory] = None
    """
    The Flop History
    """
    turn: Optional[BettingRoundHistory] = None
    """
    The Turn History
    """
    river: Optional[BettingRoundHistory] = None
    """
    The River History
    """

    def to_string(self) -> str:
        """
        Returns the string representation of the hand history, including the blind sizes,
        chips, players cards, in addition to revealed cards per street and an ordered list
        of actions per street. By convention, we assign the button an ID of 0 in this
        representation.

        Returns:
            str: The string representation of the hand history.

        """
        num_players = len(self.prehand.player_chips)
        old_ids = [i % num_players
                   for i in range(self.prehand.btn_loc, self.prehand.btn_loc + num_players)
                   if self.prehand.player_chips[i % num_players] > 0]
        canon_ids = dict(zip(old_ids, range(len(old_ids))))

        string = ""

        for history_item, name in [(self.prehand, HandPhase.PREHAND.name),
                                   (self.preflop, HandPhase.PREFLOP.name),
                                   (self.flop, HandPhase.FLOP.name),
                                   (self.turn, HandPhase.TURN.name),
                                   (self.river, HandPhase.RIVER.name),
                                   (self.settle, HandPhase.SETTLE.name)]:
            if history_item is not None:
                string += f"{name.upper()}\n" + history_item.to_string(canon_ids)
                string += '\n' if name == HandPhase.SETTLE.name else '\n\n'

        return string

    @staticmethod
    def _strip_comments(string: str) -> str:
        """
        Arguments:
            string (str): A history string
        Returns:
            str: The history string without comments.

        """
        new_lines = []
        for line in string.split('\n'):
            comment_index = line.find('#')

            if comment_index == -1:
                new_lines.append(line)
            elif comment_index != 0:
                new_lines.append(line[:comment_index].strip())

        return '\n'.join(new_lines)

    @staticmethod
    def from_string(string: str) -> History:
        """
        Reverse of :meth:`to_string()`

        Arguments:
            string (str): The string as returned from :meth:`to_string()`
        Returns:
            History: The hand history as represented by the string
        Raises:
            HistoryImportError: If the PGN file is invalid

        """
        history = History()
        sections = History._strip_comments(string).split('\n\n')

        for section in sections:
            newline = section.find('\n')
            header, rest = section[:newline], section[(newline+1):]
            if header == HandPhase.PREHAND.name:
                history_item = PrehandHistory
            elif header in (HandPhase.PREFLOP.name,
                            HandPhase.FLOP.name,
                            HandPhase.TURN.name,
                            HandPhase.RIVER.name):
                history_item = BettingRoundHistory
            elif header == HandPhase.SETTLE.name:
                # remove trailing newline for end of line
                history_item, rest = SettleHistory, rest[:-1]
            else:
                raise HistoryImportError(f"Unexpected header in history: '{header}'")

            history_item = history_item.from_string(rest)
            history[HandPhase[header]] = history_item

        return history

    def export_history(self, path: Union[str, os.PathLike] = "./texas.pgn") -> os.PathLike:
        """
        Exports the hand history to a human-readable file. If a directory is given,
        finds a name of the form :code:`texas(n).pgn` to export to. PGN files can consequently
        be imported with :meth:`import_history()`.

        Arguments:
            path (str | os.PathLike]): The directory or file to export the history to, defaults
                to the current working directory at the file `./texas.pgn`
        Returns:
            os.PathLike: The path to the history file

        """
        path_or_dir = Path(path)
        hist_path = path_or_dir

        if not hist_path.suffixes:
            hist_path.mkdir(parents=True, exist_ok=True)
            hist_path = hist_path / f"texas.{FILE_EXTENSION}"

        if f".{FILE_EXTENSION}" not in hist_path.suffixes:
            hist_path = hist_path.parent / f"{hist_path.name}.{FILE_EXTENSION}"

        # resolve lowest file_num
        original_path = hist_path
        num = 1
        while hist_path.exists():
            hist_path = original_path.parent / f"{original_path.stem}({num}).{FILE_EXTENSION}"
            num += 1

        with open(hist_path, mode="w+", encoding="ascii") as file:
            file.write(self.to_string())

        return hist_path.absolute().resolve()

    @staticmethod
    def import_history(path: Union[str, os.PathLike]) -> History:
        """
        Import a PGN file i.e. as exported from :meth:`export_history()`. Returns an
        iterator over game states as specified from the history.

        Arguments:
            path (str | os.PathLike): The PGN file to import from
        Returns:
            History: The history object represented by the PGN file
        Raises:
            HistoryImportError: If the file given does not exist or if the file is invalid

        """
        # pylint: disable=protected-access
        path = Path(path)
        if not path.exists():
            raise HistoryImportError(f'File not found: {path.absolute()}')

        # reconstitute history
        with open(path, mode='r', encoding='ascii') as file:
            try:
                history = History.from_string(file.read())
            except ValueError as err:
                raise HistoryImportError() from err

        # run checks
        history._check_missing_sections()
        history._check_unique_cards()
        history._check_correct_board_len()

        if len(history.prehand.player_chips) <= 1:
            raise HistoryImportError(f"Expected at least 2 players in this game, "
                                     f"got {len(history.prehand.player_chips)}")

        return history

    def _check_missing_sections(self):
        """
        Raises:
            HistoryImportError: If there is a section missing

        """
        if not self.preflop:
            raise HistoryImportError("Preflop section is missing")
        if not self.settle:
            raise HistoryImportError("Settle section is missing")

        for hand_phase in (HandPhase.PREFLOP,
                           HandPhase.FLOP,
                           HandPhase.TURN):
            if self[hand_phase.next_phase()] and not self[hand_phase]:
                raise HistoryImportError(f"Found a section for {hand_phase.next_phase().name} "
                                         f"but not a section for {hand_phase.name}")

    def _check_unique_cards(self):
        """
        Raises:
            HistoryImportError: If the cards in the history are not unique

        """
        cards = sum(self.prehand.player_cards.values(), [])
        for hand_phase in (HandPhase.PREFLOP,
                           HandPhase.FLOP,
                           HandPhase.TURN,
                           HandPhase.RIVER):
            history_item = self[hand_phase]
            if history_item:
                cards += history_item.new_cards

        if len(cards) != len(set(cards)):
            raise HistoryImportError("Expected cards given in history to be unique.")

    def _check_correct_board_len(self):
        """
        Raises:
            HistoryImportError: If the cards do not come out in the proper amount

        """
        total_board_len = 0
        for hand_phase in (HandPhase.PREFLOP,
                           HandPhase.FLOP,
                           HandPhase.TURN,
                           HandPhase.RIVER):
            history_item = self[hand_phase]
            if history_item:
                if len(history_item.new_cards) != hand_phase.new_cards():
                    raise HistoryImportError(f"Expected {hand_phase.new_cards()} "
                                             f"new cards in phase {hand_phase.name}")
                total_board_len += len(history_item.new_cards)

        # settle
        for _, (_, rank, _) in self.settle.pot_winners.items():
            if rank != -1:
                if len(self.settle.new_cards) != 5 - total_board_len:
                    raise HistoryImportError(f"Expected {5 - total_board_len} "
                                             f"to come out during settle phase")

    def __setitem__(self, hand_phase: HandPhase,
                    history: Union[PrehandHistory, BettingRoundHistory, SettleHistory]) -> None:
        setattr(self, hand_phase.name.lower(), history)

    def __getitem__(self, hand_phase: HandPhase) -> \
            Union[PrehandHistory, BettingRoundHistory, SettleHistory]:
        return getattr(self, hand_phase.name.lower())

    def __contains__(self, hand_phase: HandPhase) -> bool:
        if self.__getitem__(hand_phase):
            return True
        return False

    def combined_actions(self) -> list[PlayerAction]:
        """
        A list of :class:`PlayerAction` objects that was taken.

        Returns:
            list[PlayerAction]: The cumulative actions of all players in the game
        """
        history = []
        for history_item in (self.preflop,
                             self.flop,
                             self.turn,
                             self.river):
            if history_item:
                history.extend(history_item.actions)
        return history



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['texasholdem',
 'texasholdem.agents',
 'texasholdem.card',
 'texasholdem.evaluator',
 'texasholdem.game',
 'texasholdem.gui',
 'texasholdem.util']

package_data = \
{'': ['*']}

install_requires = \
['Deprecated>=1.2.13,<2.0.0']

extras_require = \
{':sys_platform == "win32"': ['windows-curses>=2.3.0,<3.0.0']}

setup_kwargs = {
    'name': 'texasholdem',
    'version': '0.7.2',
    'description': 'A texasholdem python package',
    'long_description': '# texasholdem\n![Pytest Status](https://github.com/SirRender00/texasholdem/actions/workflows/pytest.yml/badge.svg)\n[![License: MIT](https://img.shields.io/badge/License-MIT-blue.svg)](https://raw.githubusercontent.com/SirRender00/texasholdem/main/LICENSE)\n[![Documentation Status](https://readthedocs.org/projects/texasholdem/badge/?version=stable)](https://texasholdem.readthedocs.io/en/stable/?badge=stable)\n![Pylint Status](https://github.com/SirRender00/texasholdem/actions/workflows/pylint.yml/badge.svg)\n\nA python package for Texas Hold \'Em Poker providing\n- Fast evaluation of hand strengths \n- Export & import human-readable game history\n- GUIs to view games and game history\n- Simple & complex agents \n- Compliance with World Series of Poker Official Rules\n- And more\n\n| Version Name | Latest Tag | Release Notes | Patch Notes | Documentation | Release Date | End Support Date |\n| ------------ | ---------- | ------------- | ----------- | ------------- | ------------ | ---------------- |\n| 0.7          | v0.7.2     | [Release Notes](https://github.com/SirRender00/texasholdem/releases/tag/v0.7.0) | [Patch Notes](https://github.com/SirRender00/texasholdem/releases/tag/v0.7.2) | [Documentation](https://texasholdem.readthedocs.io/en/0.7/) | 16 April 2022 | |\n| 0.6          | v0.6.5     | [Release Notes](https://github.com/SirRender00/texasholdem/releases/tag/v0.6.0) | [Patch Notes](https://github.com/SirRender00/texasholdem/releases/tag/v0.6.5) | [Documentation](https://texasholdem.readthedocs.io/en/0.6/) | 24 March 2022 | |\n| 0.5          | v0.5.3     | [Release Notes](https://github.com/SirRender00/texasholdem/releases/tag/v0.5.0) | [Patch Notes](https://github.com/SirRender00/texasholdem/releases/tag/v0.5.3) | [Documentation](https://texasholdem.readthedocs.io/en/0.5/) | 21 March 2022 | |\n\nCurrent Roadmap \\\n[v1.0.0](https://github.com/SirRender00/texasholdem/wiki/Version-1.0.0-Roadmap)\n\n## Changelog v0.7\nThis release features an overhaul to the GUI system and specifically the `TextGUI`\nhad a massive overhaul.\n\n### Features\n\n- Added an `AbstractGUI` class for common functionality for all GUIs.\n- The new `TextGUI`\n    - A new history panel\n    - Support any number of players 2 thru 9\n    - Chip animations\n    - Improved UX\n\n### Other Changes\n\n- Simplification of a few steps in a betting round\n- Uncaps the python dependency\n\n## Contributing\nWant a new feature, found a bug, or have questions? Feel free to add to our issue board on Github!\n[Open Issues](https://github.com/SirRender00/texasholdem/issues>).\n\nWe welcome any developer who enjoys the package enough to contribute! Please message me at evyn.machi@gmail.com\nif you want to be added as a contributor and check out the \n[Developer\'s Guide](https://github.com/SirRender00/texasholdem/wiki/Developer\'s-Guide).\n\n## Install\nThe package is available on pypi and can be installed with\n\n```bash\npip install texasholdem\n```\n\nFor the latest experimental version\n```bash\npip install texasholdem --pre\n```\n\n## Quickstart\nPlay a game from the command line and take turns for every player out of the box.\n\n```python\nfrom texasholdem.game.game import TexasHoldEm\nfrom texasholdem.gui.text_gui import TextGUI\n\ngame = TexasHoldEm(buyin=500, big_blind=5, small_blind=2, max_players=6)\ngui = TextGUI(game=game)\n\nwhile game.is_game_running():\n    game.start_hand()\n\n    while game.is_hand_running():\n        gui.run_step()\n\n    path = game.export_history(\'./pgns\')     # save history\n    gui.replay_history(path)                 # replay history\n```\n\n## Overview\nThe following is a quick summary of what\'s in the package. Please see the \n[docs](https://texasholdem.readthedocs.io/en/stable/) for all the details.\n\n### Game Information\n\nGet game information and take actions through intuitive attributes.\n\n```python\nfrom texasholdem import TexasHoldEm, HandPhase, ActionType\n\ngame = TexasHoldEm(buyin=500,\n                   big_blind=5,\n                   small_blind=2,\n                   max_players=9)\ngame.start_hand()\n\nassert game.hand_phase == HandPhase.PREFLOP\nassert HandPhase.PREFLOP.next_phase() == HandPhase.FLOP\nassert game.chips_to_call(game.current_player) == game.big_blind\nassert len(game.get_hand(game.current_player)) == 2\n\ngame.take_action(ActionType.CALL)\n\nplayer_id = game.current_player\ngame.take_action(ActionType.RAISE, total=10)\nassert game.player_bet_amount(player_id) == 10\nassert game.chips_at_stake(player_id) == 20     # total amount in all pots the player is in\n\nassert game.chips_to_call(game.current_player) == 10 - game.big_blind\n```\n\n### Cards\nThe card module represents cards as 32-bit integers for simple and fast hand\nevaluations.\n\n```python\nfrom texasholdem import Card\n\ncard = Card("Kd")                       # King of Diamonds\nassert isinstance(card, int)            # True\nassert card.rank == 11                  # 2nd highest rank (0-12)\nassert card.pretty_string == "[ K ♦ ]"\n```\n\n### Agents\nThe package also comes with basic agents including `call_agent` and `random_agent`\n\n```python\nfrom texasholdem import TexasHoldEm\nfrom texasholdem.agents import random_agent, call_agent\n\ngame = TexasHoldEm(buyin=500, big_blind=5, small_blind=2)\ngame.start_hand()\n\nwhile game.is_hand_running():\n    if game.current_player % 2 == 0:\n        game.take_action(*random_agent(game))\n    else:\n        game.take_action(*call_agent(game))\n```\n\n### Game History\nExport and import the history of hands to files.\n\n```python\nfrom texasholdem import TexasHoldEm\nfrom texasholdem.gui import TextGUI\n\ngame = TexasHoldEm(buyin=500, big_blind=5, small_blind=2)\ngame.start_hand()\n\nwhile game.is_hand_running():\n    game.take_action(*some_strategy(game))\n\n# export to file\ngame.export_history("./pgns/my_game.pgn")\n\n# import and replay\ngui = TextGUI()\ngui.replay_history("./pgns/my_game.pgn")\n```\nPGN files also support single line and end of line comments starting with "#".\n\n### Poker Evaluator\nThe evaluator module returns the rank of the best 5-card hand from a list of 5 to 7 cards.\nThe rank is a number from 1 (strongest) to 7462 (weakest).\n\n```python\nfrom texasholdem import Card\nfrom texasholdem.evaluator import  evaluate, rank_to_string\n\nassert evaluate(cards=[Card("Kd"), Card("5d")],\n                board=[Card("Qd"),\n                       Card("6d"),\n                       Card("5s"),\n                       Card("2d"),\n                       Card("5h")]) == 927\nassert rank_to_string(927) == "Flush, King High"\n```\n\n### GUIs\nThe GUI package currently comes with a text-based GUI to play games from the command line. Coming later\nwill be web-app based GUIs.\n',
    'author': 'Evyn Machi',
    'author_email': 'evyn.machi@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/SirRender00/texasholdem',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from enum import Enum, auto


class PlayerState(Enum):
    """
    Player state Enum. For example, if a player is in the pot with the proper amount
    of chips, that player is said to be :obj:`PlayerState.IN`. If a player needs to call a bet,
    that player has status :obj:`PlayerState.TO_CALL`. If a player has no more chips to bet,
    that player is :obj:`PlayerState.ALL_IN`, etc.

    """

    SKIP = auto()
    """Player is sitting out this hand, they will not be dealt
    cards and will rejoin upon request. Will be implemented in a future version."""

    OUT = auto()
    """Player has folded their hand this round."""

    IN = auto()
    """Player is in the latest pot and has put in enough chips."""

    TO_CALL = auto()
    """Player is in the latest pot and needs to call a raise."""

    ALL_IN = auto()
    """Player is all-in and cannot take more actions this hand."""



"""The evaluator package"""

from texasholdem.evaluator.evaluator import (evaluate,
                                             get_rank_class,
                                             rank_to_string,
                                             get_five_card_rank_percentage)



"""
Evaluates hand strengths with optimizations in terms of speed and memory usage.

"""

import itertools
from typing import List

from texasholdem.card import card
from texasholdem.card.card import Card
from texasholdem.evaluator.lookup_table import LOOKUP_TABLE


def _five(cards: List[Card]) -> int:
    """
    Performs an evaluation given card in integer form, mapping them to
    a rank in the range [1, 7462], with lower ranks being more powerful.

    Variant of Cactus Kev's 5 card evaluator.

    Args:
        cards (List[Card]): A list of 5 card ints.
    Returns:
        int: The rank of the hand.

    """
    # if flush
    if cards[0] & cards[1] & cards[2] & cards[3] & cards[4] & 0xF000:
        hand_or = (cards[0] | cards[1] | cards[2] | cards[3] | cards[4]) >> 16
        prime = card.prime_product_from_rankbits(hand_or)
        return LOOKUP_TABLE.flush_lookup[prime]

    # otherwise
    prime = card.prime_product_from_hand(cards)
    return LOOKUP_TABLE.unsuited_lookup[prime]


def evaluate(cards: List[Card], board: List[Card]) -> int:
    """
    Evaluates the best five-card hand from the given cards and board. Returns
    the corresponding rank.

    Args:
        cards (List[int]): A list of length two of card ints that a player holds.
        board (List[int]): A list of length 3, 4, or 5 of card ints.
    Returns:
        int: A number between 1 (highest) and 7462 (lowest) representing the relative
            hand rank of the given card.

    """
    all_cards = cards + board
    return min(_five(hand) for hand in itertools.combinations(all_cards, 5))


def get_rank_class(hand_rank: int) -> int:
    """
    Returns the class of hand given the hand hand_rank returned from evaluate from
    9 rank classes.

    Example:
        straight flush is class 1, high card is class 9, full house is class 3.

    Returns:
        int: A rank class int describing the general category of hand from 9 rank classes.
            Example, straight flush is class 1, high card is class 9, full house is class 3.

    """
    max_rank = min(rank
                   for rank in LOOKUP_TABLE.MAX_TO_RANK_CLASS
                   if hand_rank <= rank)
    return LOOKUP_TABLE.MAX_TO_RANK_CLASS[max_rank]


def rank_to_string(hand_rank: int) -> str:
    """
    Returns a string describing the hand of the hand_rank.

    Example:
        166 -> "Four of a Kind"

    Args:
        hand_rank (int): The rank of the hand given by :meth:`evaluate`
    Returns:
        string: A human-readable string of the hand rank (i.e. Flush, Ace High).

    """
    return LOOKUP_TABLE.RANK_CLASS_TO_STRING[get_rank_class(hand_rank)]


def get_five_card_rank_percentage(hand_rank: int) -> float:
    """
    The percentage of how many of the 7462 hand strengths are worse than the given one.

    Args:
        hand_rank (int): The rank of the hand given by :meth:`evaluate`
    Returns:
        float: The percentile strength of the given hand_rank (i.e. what percent of hands is worse
            than the given one).

    """
    return 1 - float(hand_rank) / float(LOOKUP_TABLE.MAX_HIGH_CARD)
